"""Module contains Cluster business logic functions."""
import configparser
import hashlib
import logging
import os
import tempfile
import time
from datetime import datetime
from enum import Enum
from typing import Optional

import requests

from mcs_node_control.models.misc import get_dbrm_master
from mcs_node_control.models.node_config import NodeConfig
from tracing.traced_session import get_traced_session

from cmapi_server.constants import (
    CMAPI_CONF_PATH,
    CMAPI_PORT,
    DEFAULT_MCS_CONF_PATH,
    DMLPROC_SHUTDOWN_TIMEOUT,
    REQUEST_TIMEOUT,
)
from cmapi_server.exceptions import CMAPIBasicError, exc_to_cmapi_error
from cmapi_server.controllers.api_clients import NodeControllerClient
from cmapi_server.helpers import (
    broadcast_new_config,
    get_active_nodes,
    get_config_parser,
    get_current_key,
    get_dbroots,
    get_version,
    update_revision_and_manager,
)
from cmapi_server.node_manipulation import (
    add_dbroot,
    add_node,
    remove_node,
    switch_node_maintenance,
    update_dbroots_of_read_replicas,
)


class ClusterAction(Enum):
    START = 'start'
    STOP = 'stop'


def toggle_cluster_state(
        action: ClusterAction, config: str, timeout: int = DMLPROC_SHUTDOWN_TIMEOUT) -> dict:
    """Toggle the state of the cluster (start or stop).

    :param action: The cluster action to perform.
                   (ClusterAction.START or ClusterAction.STOP).
    :type action: ClusterAction
    :param config: The path to the MariaDB Columnstore configuration file.
    :type config: str
    """
    if action == ClusterAction.START:
        maintainance_flag = False
    elif action == ClusterAction.STOP:
        maintainance_flag = True
    else:
        raise ValueError(
            'Invalid action. Use ClusterAction.START or ClusterAction.STOP.'
        )

    switch_node_maintenance(maintainance_flag)
    update_revision_and_manager()
    broadcast_new_config(config, distribute_secrets=True, timeout=timeout)


class ClusterHandler:
    """Class for handling MCS Cluster operations."""

    @staticmethod
    def status(config: str = DEFAULT_MCS_CONF_PATH) -> dict:
        """Method to get MCS Cluster status information

        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :raises CMAPIBasicError: if catch some exception while getting status
                                 from each node separately
        :return: status result
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug('Cluster status command called. Getting status.')

        response = {'timestamp': str(datetime.now())}
        active_nodes = get_active_nodes(config)
        cmapi_cfg_parser = get_config_parser(CMAPI_CONF_PATH)
        api_key = get_current_key(cmapi_cfg_parser)
        headers = {'x-api-key': api_key}
        num_nodes = 0

        for node in active_nodes:
            url = f'https://{node}:8640/cmapi/{get_version()}/node/status'
            try:
                r = get_traced_session().request(
                    'GET', url, verify=False, headers=headers, timeout=REQUEST_TIMEOUT
                )
                r.raise_for_status()
                r_json = r.json()
                if len(r_json.get('services', 0)) == 0:
                    r_json['dbrm_mode'] = 'offline'
                    r_json['cluster_mode'] = 'offline'
                # add node state field ('online' if services not empty and mode not offline)
                services = r_json.get('services', [])
                node_state = (
                    'offline'
                    if not services or r_json.get('cluster_mode') == 'offline'
                    else 'online'
                )
                r_json['state'] = node_state
                response[f'{str(node)}'] = r_json
                num_nodes += 1
            except (requests.exceptions.RequestException, ValueError) as err:
                # Do not fail the whole request: record node as unreachable
                logger.error('Error retrieving status from node %s: %s', node, str(err))
                try:
                    node_dbroots = sorted(get_dbroots(node, config))
                except Exception as e:
                    logger.warning(
                        'ClusterHandler.status: failed to obtain dbroots for node %s: %s. Using empty list.',
                        node, e
                    )
                    node_dbroots = []
                response[str(node)] = {
                    'timestamp': str(datetime.now()),
                    'uptime': None,
                    'dbrm_mode': 'offline',
                    'cluster_mode': 'offline',
                    'dbroots': node_dbroots,
                    'module_id': 0,
                    'services': [],
                    'state': 'offline',
                    'error': f'Unreachable: {err.__class__.__name__}'
                }

        # num_nodes stays as number of reachable nodes
        response['num_nodes'] = num_nodes
        logger.debug('Successfully finished getting cluster status.')
        return response

    @staticmethod
    def start(config: str = DEFAULT_MCS_CONF_PATH) -> dict:
        """Method to start MCS Cluster.

        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :raises CMAPIBasicError: if no nodes in the cluster
        :return: start timestamp
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.info('Cluster start command called. Starting the cluster.')
        operation_start_time = str(datetime.now())
        toggle_cluster_state(ClusterAction.START, config)
        logger.info('Successfully finished cluster start.')
        return {'timestamp': operation_start_time}

    @staticmethod
    def shutdown(
        config: str = DEFAULT_MCS_CONF_PATH, timeout: int = DMLPROC_SHUTDOWN_TIMEOUT,
    ) -> dict:
        """Method to stop the MCS Cluster.

        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :param timeout: timeout in seconds to gracefully stop DMLProc,
                        defaults to DMLPROC_SHUTDOWN_TIMEOUT
        :type timeout: Optional[int], optional
        :raises CMAPIBasicError: if no nodes in the cluster
        :return: start timestamp
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug(
            'Cluster shutdown command called. Shutting down the cluster.'
        )
        operation_start_time = str(datetime.now())
        toggle_cluster_state(ClusterAction.STOP, config, timeout=timeout)
        logger.debug('Successfully finished shutting down the cluster.')
        return {'timestamp': operation_start_time}

    @staticmethod
    def add_node(
        node: str, config: str = DEFAULT_MCS_CONF_PATH,
        read_replica: bool = False,
    ) -> dict:
        """Method to add node to MCS CLuster.

        :param node: node IP or name or FQDN
        :type node: str
        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :param read_replica: add node as read replica, defaults to False
        :type read_replica: bool, optional
        :raises CMAPIBasicError: on exception while starting transaction
        :raises CMAPIBasicError: if transaction start isn't successful
        :raises CMAPIBasicError: on exception while adding node
        :raises CMAPIBasicError: on exception while distributing new config
        :raises CMAPIBasicError: on unsuccessful distibuting config file
        :raises CMAPIBasicError: on exception while committing transaction
        :return: result of adding node
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug(
            f'Cluster add node command called. Adding node {node} in '
            f'{"read-replica" if read_replica else "read-write"} mode.'
        )

        response = {'timestamp': str(datetime.now())}

        with exc_to_cmapi_error(prefix='Error while adding node'):
            add_node(
                node, input_config_filename=config,
                output_config_filename=config,
                read_replica=read_replica,
            )
            if not get_dbroots(node, config):
                if not read_replica:  # Read replicas don't own dbroots
                    add_dbroot(
                        host=node, input_config_filename=config,
                        output_config_filename=config
                    )

        response['node_id'] = node
        update_revision_and_manager(
            input_config_filename=config, output_config_filename=config
        )
        broadcast_new_config(config, distribute_secrets=True)
        ClusterHandler.check_shared_storage()
        logger.debug(f'Successfully finished adding node {node}.')
        return response

    @staticmethod
    def remove_node(node: str, config: str = DEFAULT_MCS_CONF_PATH) -> dict:
        """Method to remove node from MCS CLuster.

        :param node: node IP or name or FQDN
        :type node: str
        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :raises CMAPIBasicError: on exception while starting transaction
        :raises CMAPIBasicError: if transaction start isn't successful
        :raises CMAPIBasicError: on exception while removing node
        :raises CMAPIBasicError: on exception while distributing new config
        :raises CMAPIBasicError: on unsuccessful distibuting config file
        :raises CMAPIBasicError: on exception while committing transaction
        :return: result of node removing
        :rtype: dict
        """
        #TODO: This method will be moved to transaction manager in next release
        #      Due to specific use of txn_nodes inside.
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug(
            f'Cluster remove node command called. Removing node {node}.'
        )
        response = {'timestamp': str(datetime.now())}

        with exc_to_cmapi_error(prefix='Error while removing node'):
            remove_node(
                node, input_config_filename=config,
                output_config_filename=config
            )

        response['node_id'] = node
        active_nodes = get_active_nodes(config)
        if len(active_nodes) > 0:
            with NodeConfig().modify_config(config) as root:
                update_dbroots_of_read_replicas(root)

            update_revision_and_manager(
                input_config_filename=config, output_config_filename=config
            )
            broadcast_new_config(config, nodes=active_nodes)
        ClusterHandler.check_shared_storage()
        logger.debug(f'Successfully finished removing node {node}.')
        return response

    @staticmethod
    def set_mode(
        mode: str, timeout: int = 60, config: str = DEFAULT_MCS_CONF_PATH,
    ) -> dict:
        """Method to set MCS CLuster mode.

        :param mode: cluster mode to set, can be only "readonly" or "readwrite"
        :type mode: str
        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :raises CMAPIBasicError: if no master found in the cluster
        :raises CMAPIBasicError: on exception while starting transaction
        :raises CMAPIBasicError: if transaction start isn't successful
        :raises CMAPIBasicError: on exception while adding node
        :raises CMAPIBasicError: on exception while distributing new config
        :raises CMAPIBasicError: on unsuccessful distibuting config file
        :raises CMAPIBasicError: on exception while committing transaction
        :return: result of adding node
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug(
            f'Cluster mode set command called. Setting mode to {mode}.'
        )

        response = {'timestamp': str(datetime.now())}
        cmapi_cfg_parser = get_config_parser(CMAPI_CONF_PATH)
        api_key = get_current_key(cmapi_cfg_parser)
        headers = {'x-api-key': api_key}

        master = None
        if len(get_active_nodes(config)) != 0:
            master = get_dbrm_master(config)

        if master is None:
            raise CMAPIBasicError('No master found in the cluster.')
        else:
            master = master['IPAddr']
            payload: dict = {}
            payload['cluster_mode'] = mode
            url = f'https://{master}:8640/cmapi/{get_version()}/node/config'

        nc = NodeConfig()
        root = nc.get_current_config_root(config_filename=config)
        payload['manager'] = root.find('./ClusterManager').text
        payload['revision'] = root.find('./ConfigRevision').text
        payload['timeout'] = timeout
        payload['cluster_mode'] = mode

        try:
            r = get_traced_session().request(
                'PUT', url, headers=headers, json=payload, verify=False
            )
            r.raise_for_status()
            response['cluster-mode'] = mode
        except Exception as err:
            raise CMAPIBasicError(
                f'Error while setting cluster mode to {mode}'
            ) from err

        logger.debug(f'Successfully set cluster mode to {mode}.')
        return response

    @staticmethod
    def set_api_key(
        api_key: str, verification_key: str,
        config: str = DEFAULT_MCS_CONF_PATH,
    ) -> dict:
        """Method to set API key for each CMAPI node in cluster.

        :param api_key: API key to set
        :type api_key: str
        :param verification_key: TOTP key to verify
        :type verification_key: str
        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :raises CMAPIBasicError: if catch some exception while setting API key
                                 to each node
        :return: status result
        :rtype: dict
        """
        logger: logging.Logger = logging.getLogger('cmapi_server')
        logger.debug('Cluster set API key command called.')

        active_nodes = get_active_nodes(config)
        body = {
            'api_key': api_key,
            'verification_key': verification_key
        }
        response = {}
        # only for changing response object below
        active_nodes_count = len(active_nodes)

        if not active_nodes:
            # set api key in configuration file on this node
            logger.debug(
                'No active nodes found, set API key into current CMAPI conf.'
            )
            active_nodes.append('localhost')

        for node in active_nodes:
            logger.debug(f'Setting new api key to "{node}".')
            url = f'https://{node}:8640/cmapi/{get_version()}/node/apikey-set'
            try:
                resp = get_traced_session().request('PUT', url, verify=False, json=body, headers={})
                resp.raise_for_status()
                r_json = resp.json()
                if active_nodes_count > 0:
                    response[str(node)] = r_json
            except Exception as err:
                raise CMAPIBasicError(
                    f'Got an error setting API key to "{node}".'
                ) from err
            logger.debug(f'Successfully set new api key to "{node}".')

        response['timestamp'] = str(datetime.now())
        logger.debug(
            'Successfully finished setting new API key to all nodes.'
        )
        return response

    @staticmethod
    def set_log_level(
        level: str, config: str = DEFAULT_MCS_CONF_PATH,
        logger: logging.Logger = logging.getLogger('cmapi_server')
    ) -> dict:
        """Method to set level for loggers on each CMAPI node in cluster.

        :param level: logging level, including custom
        :type level: str
        :param config: columnstore xml config file path,
                       defaults to DEFAULT_MCS_CONF_PATH
        :type config: str, optional
        :param logger: logger, defaults to logging.getLogger('cmapi_server')
        :type logger: logging.Logger, optional
        :return: status result
        :rtype: dict
        """
        logger.debug('Cluster set new logging level called.')

        active_nodes = get_active_nodes(config)
        body = {'level': level}
        response = {}
        # only for changing response object below
        active_nodes_count = len(active_nodes)

        if not active_nodes:
            # set api key in configuration file on this node
            logger.debug(
                'No active nodes found, set log level onluy for current node.'
            )
            active_nodes.append('localhost')

        for node in active_nodes:
            logger.debug(f'Setting new log level to "{node}".')
            url = f'https://{node}:8640/cmapi/{get_version()}/node/log-level'
            try:
                resp = get_traced_session().request('PUT', url, verify=False, json=body, headers={})
                resp.raise_for_status()
                r_json = resp.json()
                if active_nodes_count > 0:
                    response[str(node)] = r_json
            except Exception as err:
                raise CMAPIBasicError(
                    f'Got an error setting log level to "{node}".'
                ) from err
            logger.debug(f'Successfully set new log level to "{node}".')

        response['timestamp'] = str(datetime.now())
        logger.debug(
            'Successfully finished setting new log level to all nodes.'
        )
        return response

    @staticmethod
    def check_shared_storage(skip_nodes: Optional[list[str]] = None) -> dict:
        """Check shared storage.

        :return: status result
        """
        tmp_file_path: str
        logger = logging.getLogger('shared_storage_monitor')
        active_nodes = get_active_nodes()
        if skip_nodes:
            # Remove any nodes the caller asked us to skip (e.g., unstable HB)
            active_nodes = [n for n in active_nodes if n not in set(skip_nodes)]
        all_responses: dict = dict()
        nodes_errors: dict = dict()
        sm_parser = configparser.ConfigParser()
        sm_config_str = NodeConfig().get_current_sm_config()
        sm_parser.read_string(sm_config_str)
        storage_type = sm_parser.get(
            'ObjectStorage', 'service', fallback='LocalStorage'
        )
        file_dir = '/var/lib/columnstore/data1'
        if storage_type.lower() == 's3':
            file_dir = '/var/lib/columnstore/storagemanager/metadata/data1'

        with tempfile.NamedTemporaryFile(
            mode='wb+', delete=True, dir=file_dir, prefix='mcs_test_shared'
        ) as temp_file:
            file_data = rb'File to check shared storage working.'
            temp_file.write(file_data)
            # Make sure data is on disk/visible to other nodes before checks
            temp_file.flush()
            os.fsync(temp_file.fileno())
            tmp_file_path = temp_file.name
            logger.debug(f'Temporary file to check shared storage created at: {tmp_file_path}')
            tmp_file_md5 = hashlib.md5(file_data).hexdigest()
            logger.debug(f'Temporary file md5: {tmp_file_md5}')
            for node in active_nodes:
                logger.debug(f'Checking shared file on {node!r}.')
                client = NodeControllerClient(
                    request_timeout=REQUEST_TIMEOUT,
                    base_url=f'https://{node}:{CMAPI_PORT}'
                )
                last_err_msg = None
                for attempt in range(2):
                    try:
                        node_response = client.check_shared_file(
                            file_path=tmp_file_path, check_sum=tmp_file_md5
                        )
                        logger.debug(f'Finished checking file on {node!r}')
                        all_responses[node] = node_response
                        break
                    except CMAPIBasicError as err:  # per-node failure must not abort the whole check
                        last_err_msg = err.message
                        if attempt == 0:
                            time.sleep(1)
                        continue
                else:
                    # Retries exhausted
                    logger.warning(
                        f'Error checking shared file on {node!r}: {last_err_msg}',
                        exc_info=True
                    )
                    nodes_errors[node] = last_err_msg or 'unknown error'

        nodes_success_responses = [
            v.get('success', False) for v in all_responses.values()
        ]
        if nodes_success_responses:
            shared_storage = all(nodes_success_responses)
        else:
            # no nodes in cluster case
            shared_storage = False
        # Consider partial failures either when not all successful among reachable
        # or when some nodes were unreachable (nodes_errors present).
        partially_failed = False
        if len(active_nodes) > 2:
            if nodes_errors:
                partially_failed = True
            elif nodes_success_responses and not all(nodes_success_responses):
                partially_failed = True

        response = {
            'timestamp': str(datetime.now()),
            'shared_storage': shared_storage,
            'partially_failed': partially_failed,
            'active_nodes_count': len(active_nodes),
            'nodes_responses': {**all_responses},
            'nodes_errors': {**nodes_errors}
        }

        logger.debug(
            'Successfully finished checking shared storage on all nodes.'
        )
        return response



from array import array
from threading import Lock

# for tracking the history of heartbeat responses

class InvalidNode:
    pass

class HBHistory:
    # consts to denote state of the responses
    NoResponse = 1
    GoodResponse = 2
    LateResponse = -1
    NewNode = 0

    # By default, keep a 600 heartbeat history for each node (10 mins @ 1hb/s)
    # and consider a response late if it arrives 3+ ticks late.  3 is an arbitrary small value.
    def __init__(self, tickWindow=600, lateWindow=3):
        # a list of a heartbeats for each node.  index = str, value = array of int,
        # history flushes each time threaad restarted
        self.nodeHistory = {}
        # current tick resets to zero each time thread restarted
        self.currentTick = 0
        self.lateWindow = lateWindow
        self.mutex = Lock()
        self.tickWindow = tickWindow

    def _initNode(self, node, defaultValue = GoodResponse):
        self.nodeHistory[node] = array(
            'b', [ defaultValue for _ in range(self.tickWindow) ]
        )

    def removeNode(self, node):
        self.mutex.acquire()
        if node in self.nodeHistory:
            del self.nodeHistory[node]
        self.mutex.release()

    def keepOnlyTheseNodes(self, nodes):
        self.mutex.acquire()
        nodesToKeep = set(nodes)
        historicalNodes = set(self.nodeHistory.keys())
        for node in historicalNodes:
            if node not in nodesToKeep:
                del self.nodeHistory[node]
        self.mutex.release()

    def setCurrentTick(self, tick):
        self.mutex.acquire()

        self.currentTick = tick
        for pongs in self.nodeHistory.values():
            pongs[tick % self.tickWindow] = self.NoResponse

        self.mutex.release()

    def gotHeartbeat(self, node, tickID):
        if tickID <= self.currentTick - self.lateWindow:
            status = self.LateResponse
        else:
            status = self.GoodResponse

        self.mutex.acquire()
        if node not in self.nodeHistory:
            self._initNode(node)
        self.nodeHistory[node][tickID % self.tickWindow] = status
        self.mutex.release()

    # defaultValue is used to init a fake history for a node this code is learning about
    # 'now'.  If a node is inserted into the active list, we do not want to remove
    # it right away b/c it hasn't responded to any pings yet.  Likewise,
    # if a node is inserted into the inactive list, we do not want to activate it
    # right away b/c it has responded to all pings sent so far (0).  TBD if we want
    # to add logic to handle an 'init' value in the history.
    def getNodeHistory(self, node, tickInterval, defaultValue = GoodResponse):
        self.mutex.acquire()
        if node not in self.nodeHistory:
            self._initNode(node, defaultValue = defaultValue)

        # We don't want to return values in the range where we are likely to be
        # gathering responses.
        # The return value is the range of heartbeat responses from node from
        #    tickInterval + lateWindow ticks ago to lateWindow ticks ago

        lastIndex = (self.currentTick - self.lateWindow) % self.tickWindow
        firstIndex = lastIndex - tickInterval
        history = self.nodeHistory[node]
        if firstIndex < 0:
            ret = history[firstIndex:]
            ret.extend(history[:lastIndex])
        else:
            ret = history[firstIndex:lastIndex]

        self.mutex.release()
        return ret



import concurrent.futures
import json
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Optional

from fabric import Connection

from dataclasses import dataclass, field

@dataclass
class RemoteHost:
    """Keeps configuration for a host in the cluster"""
    name: str
    private_ip: str
    public_fqdn: str
    key_file_path: Path = field(repr=False)
    ssh_user: str = field(repr=False)

    @contextmanager
    def ssh_connection(self):
        """Connect to the host using SSH (via Fabric)"""
        conn = Connection(
            host=self.public_fqdn,
            user=self.ssh_user,
            connect_kwargs={
                "key_filename": str(self.key_file_path),
            }
        )

        yield conn

        conn.close()

    def is_reachable(self) -> bool:
        """Check if the host is reachable via SSH"""
        try:
            with self.ssh_connection() as conn:
                conn.run("echo 'SSH connection successful'", hide=True)
            return True
        except Exception as e:
            print(f"SSH connection failed for {self.name}: {e}")
            return False

    def exec_mcs(self, params: str, drop_timestamp: bool = True) -> dict:
        """Execute an mcs subcommand and return its parsed output"""
        with self.ssh_connection() as conn:
            print(f"{self.name}: Executing 'mcs {params}'")
            result = conn.sudo(f"mcs {params}", hide=True)
            if result.return_code != 0:
                raise RuntimeError(f"Failed to execute mcs command on {self.name}: {result.stderr}")
            output = json.loads(result.stdout)
            return output


class ClusterConfig:
    """Keeps configuration of the cluster"""
    def __init__(self, mcs_hosts: list[RemoteHost], maxscale_hosts: list[RemoteHost]) -> None:
        self.mcs_hosts = mcs_hosts
        self.maxscale_hosts = maxscale_hosts

    def __repr__(self) -> str:
        return f"ClusterConfig(mcs_hosts={self.mcs_hosts}"

    @property
    def primary(self) -> RemoteHost:
        return self.mcs_hosts[0]

    @property
    def replicas(self) -> list[RemoteHost]:
        return self.mcs_hosts[1:]


def run_on_all_hosts_parallel(
      hosts: list[RemoteHost],
      func: Callable[[RemoteHost], Any],
      timeout: Optional[float] = None,
      max_workers: Optional[int] = None,
    ) -> dict[str, Any]:
    """
    Run a function on all hosts in parallel.

    :param hosts: List of RemoteHost objects
    :param func: Function that takes a RemoteHost as its only argument
    :param timeout: Timeout for the function execution
    :param max_workers: Maximum number of worker threads (None = default based on system)
    :returns: Dictionary mapping hosts to function results
    """
    results = {}

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Start the function execution on each host
        future_to_host = {executor.submit(func, host): host for host in hosts}

        # Process results as they complete
        for future in concurrent.futures.as_completed(future_to_host):
            host = future_to_host[future]
            try:
                result = future.result(timeout=timeout)
                results[host.name] = result
            except Exception as exc:
                print(f"Error on host {host.name}: {exc}")
                results[host.name] = exc
                raise

    return results


def block_port(host: RemoteHost, port: int) -> None:
    """Block a port on the host using iptables"""
    with host.ssh_connection() as conn:
        print(f"{host.name}: Blocking port {port}")
        conn.sudo(f"iptables -A INPUT -p tcp --dport {port} -j DROP")
        print(f"{host.name}: Port {port} blocked")


def unblock_port(host: RemoteHost, port: int) -> None:
    """Unblock a port on the host using iptables"""
    with host.ssh_connection() as conn:
        print(f"{host.name}: Unblocking port {port}")
        conn.sudo(f"iptables -D INPUT -p tcp --dport {port} -j DROP")
        print(f"{host.name}: Port {port} unblocked")


from .. import config
import time
from socket import *
import struct
import sys

_config = config.Config("failover/test/config-test.xml")
print("got desired_nodes = {}".format(_config.getDesiredNodes()))
print("got active_nodes = {}".format(_config.getActiveNodes()))
print("got inacive_nodes = {}".format(_config.getInactiveNodes()))
print("got all nodes = {}".format(_config.getAllNodes()))
print("got primarynode = {}".format(_config.getPrimaryNode()))
print()

from ..heartbeater import HeartBeater
from ..heartbeat_history import HBHistory

hbh = HBHistory()
hb = HeartBeater(_config, hbh)
hb.start()
sock = socket(type = SOCK_DGRAM)
sock.bind(('localhost', 12345))

# Updated heartbeater to send the reply to its own port, rather than to
# the port of the sending socket.  Need to update this.
#msg = struct.pack("4sH", hb.areYouThereMsg, 1234)
#sock.sendto(msg, ('localhost', hb.port))
#print("sent the are-you-there msg")
#(data, remote) = sock.recvfrom(6)
#(data, seq) = struct.unpack("4sH", data)
#if data == hb.yesIAmMsg:
#    print("got the yes-i-am msg, seq = {}".format(seq))
#else:
#    print("got something other than the yes-i-am-msg")

hb.stop()

#from heartbeat_history import HBHistory

#hbh = HBHistory()
hbh.setCurrentTick(0)
hbh.gotHeartbeat("node1", 0)
hbh.setCurrentTick(1)
hbh.gotHeartbeat("node2", 1)
hbh.setCurrentTick(2)
hbh.setCurrentTick(10)
hbh.gotHeartbeat("node1", 9)
hbh.gotHeartbeat("node1", 2)
pongs = hbh.getNodeHistory("node1", 20)
print("Got pongs: {}".format(pongs))

print('''
 This is currently a 'manual' test, meaning the user should watch for the expected output
 In this case, because NM's identity checker will return 'node1', and that does not match
 node[2-4], those nodes will appear to NodeMonitor to be offline.  Our starting condition
 is that nodes 1-3 are active, and node4 is inactive.  After 15s, nodes 2 & 3
 should be deactivated, a new primary node will be chosen, and and our AgentBase will start 
 printing these events.
''')
def testNodeMonitor1(nm):
    nm.start()
    print("Waiting for 20 secs, watch for output from AgentBase")
    time.sleep(20)
    nm.stop()
    time.sleep(1)
    print("NodeMonitor was stopped, did it produce the right output?")

from ..node_monitor import NodeMonitor
nm = NodeMonitor(config = _config, samplingInterval = 10)
# check whether node[1-4] are in the /etc/hosts file as localhost
addr1 = gethostbyname("node1")
addr2 = gethostbyname("node2")
addr3 = gethostbyname("node3")
addr4 = gethostbyname("node4")
if addr1 == '127.0.0.1' and addr2 == '127.0.0.1' and addr3 == '127.0.0.1' and addr4 == '127.0.0.1':
    testNodeMonitor1(nm)
else:
    print("Skipping testNodeMonitor1().  node[1-4] needs to be defined as 127.0.0.1 in /etc/hosts")



print("tester is finished")




"""Module related to CMAPI transaction management logic."""
import logging
from contextlib import ContextDecorator
from signal import (
    SIGINT, SIGTERM, SIGHUP, SIG_DFL, signal, default_int_handler
)
from typing import Optional

from cmapi_server.constants import DEFAULT_MCS_CONF_PATH, TRANSACTION_TIMEOUT
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.helpers import (
    get_id, commit_transaction, rollback_transaction, start_transaction
)


class TransactionManager(ContextDecorator):
    """Context manager and decorator to put any code inside CMAPI transaction.

    :param timeout: time in sec after transaction will be autocommitted,
                    defaults to 300.0 (TRANSACTION_TIMEOUT)

    :param timeout: _description_, defaults to 300
    :type timeout: float, optional
    :param txn_id: custom transaction id, defaults to None
    :type txn_id: Optional[int], optional
    :param handle_signals: handle specific signals or not, defaults to False
    :type handle_signals: bool, optional
    :param extra_nodes: extra nodes to start transaction at, defaults to None
    :type extra_nodes: Optional[list], optional
    :param remove_nodes: nodes to remove from transaction, defaults to None
    :type remove_nodes: Optional[list], optional
    :param optional_nodes: nodes to add to transaction, defaults to None
    :type optional_nodes: Optional[list], optional
    :raises CMAPIBasicError: if there are no nodes in the cluster
    :raises CMAPIBasicError: if starting transaction isn't succesful
    :raises Exception: if error while starting the transaction
    :raises Exception: if error while committing transaction
    :raises Exception: if error while rollback transaction
    """

    def __init__(
            self, timeout: float = TRANSACTION_TIMEOUT,
            txn_id: Optional[int] = None, handle_signals: bool = False,
            extra_nodes: Optional[list] = None,
            remove_nodes: Optional[list] = None,
            optional_nodes: Optional[list] = None,
    ):
        self.timeout = timeout
        self.txn_id = txn_id or get_id()
        self.handle_signals = handle_signals
        self.active_transaction = False
        self.extra_nodes = extra_nodes
        self.remove_nodes = remove_nodes
        self.optional_nodes = optional_nodes
        self.success_txn_nodes = None

    def _handle_exception(
            self, exc: Optional[Exception] = None,
            signum: Optional[int] = None
    ) -> None:
        """Handle raised exceptions.

        We need to rollback transaction in some cases and return back default
        signal handlers.

        :param exc: exception passed, defaults to None
        :type exc: Optional[Type[Exception]], optional
        :param signum: signal if it cause exception, defaults to None
        :type signum: Optional[int], optional
        :raises exc: raises passed exception
        """
        # message = 'Got exception in transaction manager'
        if (exc or signum) and self.active_transaction:
            self.rollback_transaction(nodes=self.success_txn_nodes)
        self.set_default_signals()
        if exc is not None:
            raise exc

    def _handle_signal(self, signum, frame) -> None:
        """Handler for signals.

        :param signum: signal number
        :type signum: int
        """
        logging.error(f'Caught signal "{signum}" in transaction manager.')
        self._handle_exception(signum=signum)

    def set_custom_signals(self) -> None:
        """Set handlers for several signals."""
        # register handler for signals for proper handling them
        for sig in SIGINT, SIGTERM, SIGHUP:
            signal(sig, self._handle_signal)

    def set_default_signals(self) -> None:
        """Return defalt handlers for specific signals."""
        if self.handle_signals:
            signal(SIGINT, default_int_handler)
            signal(SIGTERM, SIG_DFL)
            signal(SIGHUP, SIG_DFL)

    def rollback_transaction(self, nodes: Optional[list] = None) -> None:
        """Rollback transaction.

        :param nodes: nodes to rollback transaction, defaults to None
        :type nodes: Optional[list], optional
        """
        try:
            rollback_transaction(self.txn_id, nodes=nodes)
            self.active_transaction = False
            logging.debug(f'Successful rollback of transaction "{self.txn_id}".')
        except Exception:
            logging.error(
                f'Error while rolling back transaction "{self.txn_id}"',
                exc_info=True
            )

    def commit_transaction(self, nodes: Optional[list] = None) -> None:
        """Commit transaction.

        :param nodes: nodes to commit transaction, defaults to None
        :type nodes: Optional[list], optional
        """
        try:
            commit_transaction(
                self.txn_id, cs_config_filename=DEFAULT_MCS_CONF_PATH,
                nodes=nodes
            )
        except Exception:
            logging.error(f'Error while committing transaction {self.txn_id}')
            self.rollback_transaction(nodes=self.success_txn_nodes)
            self.set_default_signals()
            raise

    def __enter__(self):
        if self.handle_signals:
            self.set_custom_signals()
        suceeded, success_txn_nodes = False, []
        try:
            suceeded, _, success_txn_nodes = start_transaction(
                cs_config_filename=DEFAULT_MCS_CONF_PATH,
                extra_nodes=self.extra_nodes, remove_nodes=self.remove_nodes,
                optional_nodes=self.optional_nodes,
                txn_id=self.txn_id, timeout=self.timeout,
            )
        except Exception as exc:
            logging.error('Error while starting the transaction.')
            self._handle_exception(exc=exc)
        if not suceeded:
            self._handle_exception(
                exc=CMAPIBasicError('Starting transaction isn\'t succesful.')
            )
        if suceeded and len(success_txn_nodes) == 0:
            # corner case when deleting last node in the cluster
            # TODO: remove node mechanics potentially has a vulnerability
            #       because no transaction started for removing node.
            #       Probably in some cases rollback never works for removing
            #       node, because it never exist in success_txn_nodes.
            if not self.remove_nodes:
                self._handle_exception(
                    exc=CMAPIBasicError('There are no nodes in the cluster.')
                )
        self.active_transaction = True
        self.success_txn_nodes = success_txn_nodes
        return self

    def __exit__(self, *exc):
        if exc[0] and self.active_transaction:
            self.rollback_transaction(nodes=self.success_txn_nodes)
            self.set_default_signals()
            return False
        if self.active_transaction:
            self.commit_transaction(nodes=self.success_txn_nodes)
            self.set_default_signals()
        return True



from __future__ import annotations

import logging
import os.path
import socket
import time
from copy import deepcopy
from time import sleep

import psutil

from cmapi_server.constants import (
    ALL_MCS_PROGS,
    DMLPROC_SHUTDOWN_TIMEOUT,
    MCS_INSTALL_BIN,
    MCSProgs,
    ProgInfo,
)
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.process_dispatchers.base import BaseDispatcher
from cmapi_server.process_dispatchers.container import ContainerDispatcher
from cmapi_server.process_dispatchers.systemd import SystemdDispatcher
from mcs_node_control.models.dbrm import DBRM
from mcs_node_control.models.dbrm_socket import SOCK_TIMEOUT
from mcs_node_control.models.misc import get_workernodes
from mcs_node_control.models.process import Process

PROCESS_DISPATCHERS: dict[str, type[BaseDispatcher]] = {
    'systemd': SystemdDispatcher,
    # could be used in docker containers and OSes w/o systemd
    'container': ContainerDispatcher,
}
PRIMARY_PROGS = ('controllernode', 'DMLProc', 'DDLProc')


class MCSProcessManager:
    """Class to run process operations.

    e.g. re/-start or stop systemd services, run executable.
    """
    CONTROLLER_MAX_RETRY = 30
    mcs_progs: dict[str, ProgInfo] = {}
    mcs_version_info = None
    dispatcher_name = None
    process_dispatcher: BaseDispatcher = None

    @classmethod
    def _get_prog_name(cls, name: str) -> str:
        """Get proper service name for systemd or non-systemd installations.

        :param name: service name
        :type name: str
        :return: correct service name
        :rtype: str
        """
        if cls.dispatcher_name == 'systemd':
            prog = MCSProgs(name)
            return ALL_MCS_PROGS[prog].service_name
        return name

    @classmethod
    def _get_sorted_progs(
        cls, is_primary: bool, reverse: bool = False, is_read_replica: bool = False
    ) -> dict:
        """Get sorted services dict.

        :param is_primary: is primary node or not
        :type is_primary: bool
        :param reverse: reverse sort order, defaults to False
        :type reverse: bool, optional
        :return: dict with sorted services in correct start/stop order
        :rtype: dict
        """
        unsorted_progs: dict
        if is_primary:
            unsorted_progs = deepcopy(cls.mcs_progs)
        else:
            unsorted_progs = {
                prog_name: prog_info
                for prog_name, prog_info in cls.mcs_progs.items()
                if prog_name not in PRIMARY_PROGS
            }

        if is_read_replica:
            logging.debug('Node is a read replica, skipping WriteEngine')
            unsorted_progs.pop(
                MCSProgs.WRITE_ENGINE_SERVER, None
            )

        if reverse:
            # stop sequence builds using stop_priority property
            return dict(
                sorted(
                    unsorted_progs.items(),
                    key=lambda item: item[1].stop_priority,
                )
            )
        # start up sequence is a dict default sequence
        return unsorted_progs

    @classmethod
    def _detect_processes(cls) -> None:
        """Detect existing mcs services. Depends on MCS version."""
        if cls.mcs_progs:
            logging.warning('Mcs ProcessHandler already detected processes.')

        for prog, prog_info in ALL_MCS_PROGS.items():
            prog_name = prog.value
            if os.path.exists(os.path.join(MCS_INSTALL_BIN, prog_name)):
                cls.mcs_progs[prog_name] = prog_info

    @classmethod
    def detect(cls, dispatcher_name: str, dispatcher_path: str = None) -> None:
        """Detect mcs version info and installed processes.

        :param dispatcher_name: process dispatcher name
        :type dispatcher_name: str
        :param dispatcher_path: path to custom dispatcher,
            for next releases, defaults to None
        :type dispatcher_path: str, optional
        :raises CMAPIBasicError: if custom dispatcher path doesn't exists
        :raises CMAPIBasicError: Not implemented custom dispatcher error
        """
        cls._detect_processes()
        # detect mcs version info by processes
        if len(cls.mcs_progs) == 8:
            cls.mcs_version_info = '6.4.x and lower'
        elif len(cls.mcs_progs) == 7 and 'ExeMgr' not in cls.mcs_progs:
            cls.mcs_version_info = '22.08.x and higher'
        else:
            cls.mcs_version_info = 'Undefined'
            logging.warning(
                'MCS version haven\'t been detected properly.'
                'Please try to update your CMAPI version or contact support.'
            )
        logging.info(
            f'Detected {len(cls.mcs_progs)} MCS services. '
            f'MCS version is {cls.mcs_version_info}'
        )
        # TODO: For next releases. Do we really need custom dispatchers?
        if dispatcher_name not in PROCESS_DISPATCHERS:
            logging.warning(
                f'Custom process dispatcher with name "{dispatcher_name}" '
                f'and path "{dispatcher_path}" used.'
            )
            if not dispatcher_path or not os.path.exists(dispatcher_path):
                err_msg = 'Wrong dispatcher path in cmapi_config file.'
                logging.error(err_msg)
                raise CMAPIBasicError(err_msg)
            cls.dispatcher_name = 'custom'
            raise CMAPIBasicError('Custom dispatchers yet not implemented!')

        cls.dispatcher_name = dispatcher_name
        cls.process_dispatcher = PROCESS_DISPATCHERS[dispatcher_name]
        cls.process_dispatcher.init()

    @classmethod
    def _wait_for_workernodes(cls) -> bool:
        """Wait for workernodes processes.

        Waiting for all workernodes to come up before starting
        controllernode on a primary.

        :return: True on success
        :rtype: bool
        """
        logging.debug(
            'Waiting for all workernodes to come up before starting '
            'controllernode on a primary.'
        )
        workernodes = get_workernodes()
        attempts = cls.CONTROLLER_MAX_RETRY
        while attempts > 0 and len(workernodes) > 0:
            logging.debug(f'Waiting for "{list(workernodes)}"....{attempts}')
            start = time.monotonic()
            # creating a separate list with workernode names
            # for safe deleting items from source dict
            for name in list(workernodes):
                try:
                    sock = socket.socket(
                        socket.AF_INET, socket.SOCK_STREAM
                    )
                    sock.settimeout(SOCK_TIMEOUT)
                    sock.connect(
                        (
                            workernodes[name]['IPAddr'],
                            workernodes[name]['Port']
                        )
                    )
                except (ConnectionRefusedError, socket.timeout):
                    logging.info(
                        f'"{name}" {workernodes[name]["IPAddr"]}:'
                        f'{workernodes[name]["Port"]} not started yet.'
                    )
                else:
                    logging.debug(f'Workernode {workernodes[name]["IPAddr"]}:{workernodes[name]["Port"]} started.')
                    # delete started workernode from workernodes dict
                    del workernodes[name]
                finally:
                    sock.close()

            if workernodes:
                # We get here only if some workernodes were not accessible during this attempt
                elapsed = time.monotonic() - start
                remaining = SOCK_TIMEOUT - elapsed
                if remaining > 0:
                    logging.debug('Sleeping %.1f seconds', remaining)
                    time.sleep(remaining)

            attempts -= 1

        if workernodes:
            logging.error(
                f'Some workernodes: "{workernodes}" are not reachable after '
                f'{cls.CONTROLLER_MAX_RETRY} attempts to connect with '
                f'{SOCK_TIMEOUT} seconds timeout.'
                'Starting mcs-controllernode anyway.'
            )
            return False
        return True

    @classmethod
    def _wait_for_controllernode(cls) -> bool:
        """Waiting for controllernode to come up on a primary.

        :return: True on success
        :rtype: bool
        """
        logging.debug(
            'Waiting for controllernode to come up before starting '
            'ddlproc/dmlproc on non-primary nodes.'
        )
        attempts = cls.CONTROLLER_MAX_RETRY
        success = False
        while attempts > 0:
            start = time.monotonic()
            try:
                with DBRM():
                    # check connection
                    success = True
            except (OSError, ConnectionRefusedError, RuntimeError):
                elapsed = time.monotonic() - start
                remaining = SOCK_TIMEOUT - elapsed
                logging.info(
                    'Cannot establish connection to controllernode.'
                    f'Controller node still not started. {attempts} attempts left.'
                )
                if remaining > 0:
                    logging.debug('Sleeping %.1f seconds', remaining)
                    time.sleep(remaining)
            else:
                logging.debug('Controllernode is reachable')
                break
            attempts -= 1

        if not success:
            logging.error(
                'Controllernode is not reachable after '
                f'{cls.CONTROLLER_MAX_RETRY} attempts to connect with '
                f'{SOCK_TIMEOUT} seconds timeout.'
                'Starting mcs-dmlproc/mcs-ddlproc anyway.'
            )
            return False
        return True

    @classmethod
    def _wait_for_DMLProc_stop(cls, timeout: int = DMLPROC_SHUTDOWN_TIMEOUT) -> bool:
        """Waiting DMLProc process to stop.

        :param timeout: timeout to wait in seconds, defaults to DMLPROC_SHUTDOWN_TIMEOUT
        :type timeout: int, optional
        :return: True on success
        :rtype: bool
        """
        logging.info(f'Waiting for DMLProc to stop in {timeout} seconds')
        deadline = time.monotonic() + max(1, int(timeout))
        # Log at most every 5 seconds while polling every ~1s for responsiveness
        LOG_INTERVAL_SEC = 5.0
        next_log_at = time.monotonic()  # log immediately on first iteration

        while True:
            now = time.monotonic()
            remaining = deadline - now
            if remaining <= 0:
                break

            if not Process.check_process_alive('DMLProc'):
                logging.info('DMLProc gracefully stopped by DBRM command.')
                return True

            if now >= next_log_at:
                logging.info(
                    f'Waiting for DMLProc to stop. Seconds left ~{int(remaining)}.'
                )
                next_log_at = now + LOG_INTERVAL_SEC

            # Sleep in small increments to minimize over-wait after process exit
            sleep(min(1.0, max(0.1, remaining)))

        logging.error(
            "DMLProc didn't stop gracefully by DBRM command within "
            f"{int(timeout)} seconds. Will be stopped directly."
        )
        return False

    @classmethod
    def noop(cls, *args, **kwargs):
        """No operation. TODO: looks like useless."""
        cls.process_dispatcher.noop()

    @classmethod
    def gracefully_stop_dmlproc(cls) -> None:
        """Gracefully stop DMLProc using DBRM commands."""
        logging.info(
            'Trying to gracefully stop DMLProc using DBRM commands.'
        )
        try:
            with DBRM() as dbrm:
                dbrm.set_system_state(
                    ['SS_ROLLBACK', 'SS_SHUTDOWN_PENDING']
                )
        except (ConnectionRefusedError, RuntimeError):
            logging.error(
                'Cannot set SS_ROLLBACK and SS_SHUTDOWN_PENDING via DBRM, '
                'graceful auto stop of DMLProc failed. '
                'Try a regular stop method.'
            )
            raise

    @classmethod
    def is_service_running(cls, name: str, use_sudo: bool = True) -> bool:
        """Check if MCS process is running.

        :param name: mcs process name
        :type name: str
        :param use_sudo: use sudo or not, defaults to True
        :type use_sudo: bool, optional
        :return: True if mcs process is running, otherwise False
        :rtype: bool
        """
        return cls.process_dispatcher.is_service_running(
            cls._get_prog_name(name), use_sudo
        )


    @classmethod
    def start(cls, name: str, is_primary: bool, use_sudo: bool) -> bool:
        """Start mcs process.

        :param name: mcs process name
        :type name: str
        :param is_primary: is node primary or not
        :type is_primary: bool
        :param use_sudo: use sudo or not
        :type use_sudo: bool
        :return: True if process started successfully
        :rtype: bool
        """
        return cls.process_dispatcher.start(
            cls._get_prog_name(name), is_primary, use_sudo
        )

    @classmethod
    def stop(
        cls, name: str, is_primary: bool, use_sudo: bool, timeout: int = DMLPROC_SHUTDOWN_TIMEOUT
    ) -> bool:
        """Stop mcs process.

        :param name: mcs process name
        :type name: str
        :param is_primary: is node primary or not
        :type is_primary: bool
        :param use_sudo: use sudo or not
        :type use_sudo: bool
        :param timeout: timeout for DMLProc gracefully stop using DBRM, seconds
        :type timeout: int
        :return: True if process stopped successfully
        :rtype: bool
        """
        # TODO: do we need here force stop DMLProc as a method argument?

        if is_primary and name == 'DMLProc':
            try:
                cls.gracefully_stop_dmlproc()
            except (ConnectionRefusedError, RuntimeError):
                # stop DMLProc using regular signals or systemd
                return cls.process_dispatcher.stop(
                    cls._get_prog_name(name), is_primary, use_sudo
                )
            # DMLProc gracefully stopped using DBRM commands otherwise
            # continue with a regular stop method
            if cls._wait_for_DMLProc_stop(timeout):
                return True
        return cls.process_dispatcher.stop(
            cls._get_prog_name(name), is_primary, use_sudo
        )

    @classmethod
    def restart(cls, name: str, is_primary: bool, use_sudo: bool) -> bool:
        """Restart mcs process.

        :param name: mcs process name
        :type name: str
        :param is_primary: is node primary or not
        :type is_primary: bool
        :param use_sudo: use sudo or not
        :type use_sudo: bool
        :return: True if process restarted successfully
        :rtype: bool
        """
        return cls.process_dispatcher.restart(
            cls._get_prog_name(name), is_primary, use_sudo
        )

    @classmethod
    def get_running_mcs_procs(cls) -> list[dict]:
        """Search for mcs processes.

        The method returns PIDs of MCS services in both container or systemd
        environments.

        :return: list of dicts with name and pid of mcs process
        :rtype: list[dict]
        """
        return [
            {'name': proc.name(), 'pid': proc.pid}
            for proc in psutil.process_iter(['pid', 'name'])
            if proc.name() in cls.mcs_progs
        ]

    @classmethod
    def is_node_processes_ok(
        cls, is_primary: bool, node_stopped: bool
    ) -> bool:
        """Check if needed processes exists or not.

        :param is_primary: is node primary or not
        :type is_primary: bool
        :param node_stopped: is node stopped or started
        :type node_stopped: bool
        :return: True if there are expected value of processes, else False
        :rtype: bool

        ..NOTE: For next releases. Now only used in tests.
        """
        running_procs = cls.get_running_mcs_procs()
        if node_stopped:
            return len(running_procs) == 0
        node_progs = cls._get_sorted_progs(is_primary)
        return set(node_progs) == set(p['name'] for p in running_procs)

    @classmethod
    def start_node(
        cls,
        is_primary: bool,
        use_sudo: bool = True,
        is_read_replica: bool = False,
    ) -> None:
        """Start mcs node processes.

        :param is_primary: is node primary or not, defaults to True
        :type is_primary: bool
        :param use_sudo: use sudo or not, defaults to True
        :type use_sudo: bool, optional
        :param is_read_replica: if true, doesn't start WriteEngine
        :type is_read_replica: bool, optional
        :raises CMAPIBasicError: immediately if one mcs process not started
        """
        for prog_name in cls._get_sorted_progs(
            is_primary=is_primary,
            is_read_replica=is_read_replica,
        ):
            if (
                    cls.dispatcher_name == 'systemd'
                    and prog_name == MCSProgs.STORAGE_MANAGER
            ):
                # TODO: MCOL-5458
                logging.info(
                    f'Skip starting {prog_name} with systemd dispatcher.'
                )
                continue
            # TODO: additional error handling
            if prog_name == MCSProgs.CONTROLLER_NODE:
                cls._wait_for_workernodes()
            if prog_name in (MCSProgs.DML_PROC, MCSProgs.DDL_PROC):
                cls._wait_for_controllernode()
            if not cls.start(prog_name, is_primary, use_sudo):
                logging.error(f'Process "{prog_name}" not started properly.')
                raise CMAPIBasicError(f'Error while starting "{prog_name}".')

    @classmethod
    def stop_node(
        cls,
        is_primary: bool,
        use_sudo: bool = True,
        timeout: int = DMLPROC_SHUTDOWN_TIMEOUT,
    ):
        """Stop mcs node processes.

        :param is_primary: is node primary or not, defaults to True
        :type is_primary: bool
        :param use_sudo: use sudo or not, defaults to True
        :type use_sudo: bool, optional
        :param timeout: timeout for DMLProc gracefully stop using DBRM, seconds
        :type timeout: int
        :raises CMAPIBasicError: immediately if one mcs process not stopped
        """
        # Every time try to stop all processes no matter primary it or slave,
        # so use full available list of processes. Otherwise, it could cause
        # undefined behaviour when primary gone and then recovers (failover
        # triggered 2 times).
        for prog_name in cls._get_sorted_progs(is_primary=True, reverse=True):
            if not cls.stop(prog_name, is_primary, use_sudo, timeout=timeout):
                logging.error(f'Process "{prog_name}" not stopped properly.')
                raise CMAPIBasicError(f'Error while stopping "{prog_name}"')

    @classmethod
    def restart_node(cls, is_primary: bool, use_sudo: bool, is_read_replica: bool = False):
        """TODO: For next releases."""
        if cls.get_running_mcs_procs():
            cls.stop_node(is_primary, use_sudo)
        cls.start_node(is_primary, use_sudo, is_read_replica)


class MDBProcessManager:
    """TODO: not working with a non systemd installations need to implement."""
    process_dispatcher = SystemdDispatcher

    @classmethod
    def is_service_running(cls, use_sudo: bool = True) -> bool:
        """Check if MariaDB process is running.

        :param use_sudo: use sudo or not, defaults to True
        :type use_sudo: bool, optional
        :return: True if MariaDB process is running, otherwise False
        :rtype: bool
        """
        return cls.process_dispatcher.is_service_running(
            'mariadb', use_sudo=use_sudo
        )

    @classmethod
    def start(cls, use_sudo: bool) -> bool:
        """Start MariaDB process.

        :type use_sudo: bool
        :return: True if process started successfully
        :rtype: bool
        """
        if not cls.is_service_running():
            return cls.process_dispatcher.start('mariadb', use_sudo)
        return True

    @classmethod
    def stop(cls, use_sudo: bool = True) -> bool:
        """Stop MariaDB process.

        :type use_sudo: bool
        :return: True if process stopped
          successfully
        :rtype: bool
        """
        if cls.is_service_running():
            return cls.process_dispatcher.stop('mariadb', use_sudo)
        return True



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

from abc import ABC, abstractmethod
from advisor.db_log_parser import DataSource, NO_COL_FAMILY
from advisor.db_timeseries_parser import TimeSeriesData
from enum import Enum
from advisor.ini_parser import IniParser
import re


class Section(ABC):
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def set_parameter(self, key, value):
        pass

    @abstractmethod
    def perform_checks(self):
        pass


class Rule(Section):
    def __init__(self, name):
        super().__init__(name)
        self.conditions = None
        self.suggestions = None
        self.overlap_time_seconds = None
        self.trigger_entities = None
        self.trigger_column_families = None

    def set_parameter(self, key, value):
        # If the Rule is associated with a single suggestion/condition, then
        # value will be a string and not a list. Hence, convert it to a single
        # element list before storing it in self.suggestions or
        # self.conditions.
        if key == 'conditions':
            if isinstance(value, str):
                self.conditions = [value]
            else:
                self.conditions = value
        elif key == 'suggestions':
            if isinstance(value, str):
                self.suggestions = [value]
            else:
                self.suggestions = value
        elif key == 'overlap_time_period':
            self.overlap_time_seconds = value

    def get_suggestions(self):
        return self.suggestions

    def perform_checks(self):
        if not self.conditions or len(self.conditions) < 1:
            raise ValueError(
                self.name + ': rule must have at least one condition'
            )
        if not self.suggestions or len(self.suggestions) < 1:
            raise ValueError(
                self.name + ': rule must have at least one suggestion'
            )
        if self.overlap_time_seconds:
            if len(self.conditions) != 2:
                raise ValueError(
                    self.name + ": rule must be associated with 2 conditions\
                    in order to check for a time dependency between them"
                )
            time_format = '^\d+[s|m|h|d]$'
            if (
                not
                re.match(time_format, self.overlap_time_seconds, re.IGNORECASE)
            ):
                raise ValueError(
                    self.name + ": overlap_time_seconds format: \d+[s|m|h|d]"
                )
            else:  # convert to seconds
                in_seconds = int(self.overlap_time_seconds[:-1])
                if self.overlap_time_seconds[-1] == 'm':
                    in_seconds *= 60
                elif self.overlap_time_seconds[-1] == 'h':
                    in_seconds *= (60 * 60)
                elif self.overlap_time_seconds[-1] == 'd':
                    in_seconds *= (24 * 60 * 60)
                self.overlap_time_seconds = in_seconds

    def get_overlap_timestamps(self, key1_trigger_epochs, key2_trigger_epochs):
        # this method takes in 2 timeseries i.e. timestamps at which the
        # rule's 2 TIME_SERIES conditions were triggered and it finds
        # (if present) the first pair of timestamps at which the 2 conditions
        # were triggered within 'overlap_time_seconds' of each other
        key1_lower_bounds = [
            epoch - self.overlap_time_seconds
            for epoch in key1_trigger_epochs
        ]
        key1_lower_bounds.sort()
        key2_trigger_epochs.sort()
        trigger_ix = 0
        overlap_pair = None
        for key1_lb in key1_lower_bounds:
            while (
                key2_trigger_epochs[trigger_ix] < key1_lb and
                trigger_ix < len(key2_trigger_epochs)
            ):
                trigger_ix += 1
            if trigger_ix >= len(key2_trigger_epochs):
                break
            if (
                key2_trigger_epochs[trigger_ix] <=
                key1_lb + (2 * self.overlap_time_seconds)
            ):
                overlap_pair = (
                    key2_trigger_epochs[trigger_ix],
                    key1_lb + self.overlap_time_seconds
                )
                break
        return overlap_pair

    def get_trigger_entities(self):
        return self.trigger_entities

    def get_trigger_column_families(self):
        return self.trigger_column_families

    def is_triggered(self, conditions_dict, column_families):
        if self.overlap_time_seconds:
            condition1 = conditions_dict[self.conditions[0]]
            condition2 = conditions_dict[self.conditions[1]]
            if not (
                condition1.get_data_source() is DataSource.Type.TIME_SERIES and
                condition2.get_data_source() is DataSource.Type.TIME_SERIES
            ):
                raise ValueError(self.name + ': need 2 timeseries conditions')

            map1 = condition1.get_trigger()
            map2 = condition2.get_trigger()
            if not (map1 and map2):
                return False

            self.trigger_entities = {}
            is_triggered = False
            entity_intersection = (
                set(map1.keys()).intersection(set(map2.keys()))
            )
            for entity in entity_intersection:
                overlap_timestamps_pair = (
                    self.get_overlap_timestamps(
                        list(map1[entity].keys()), list(map2[entity].keys())
                    )
                )
                if overlap_timestamps_pair:
                    self.trigger_entities[entity] = overlap_timestamps_pair
                    is_triggered = True
            if is_triggered:
                self.trigger_column_families = set(column_families)
            return is_triggered
        else:
            all_conditions_triggered = True
            self.trigger_column_families = set(column_families)
            for cond_name in self.conditions:
                cond = conditions_dict[cond_name]
                if not cond.get_trigger():
                    all_conditions_triggered = False
                    break
                if (
                    cond.get_data_source() is DataSource.Type.LOG or
                    cond.get_data_source() is DataSource.Type.DB_OPTIONS
                ):
                    cond_col_fam = set(cond.get_trigger().keys())
                    if NO_COL_FAMILY in cond_col_fam:
                        cond_col_fam = set(column_families)
                    self.trigger_column_families = (
                        self.trigger_column_families.intersection(cond_col_fam)
                    )
                elif cond.get_data_source() is DataSource.Type.TIME_SERIES:
                    cond_entities = set(cond.get_trigger().keys())
                    if self.trigger_entities is None:
                        self.trigger_entities = cond_entities
                    else:
                        self.trigger_entities = (
                            self.trigger_entities.intersection(cond_entities)
                        )
                if not (self.trigger_entities or self.trigger_column_families):
                    all_conditions_triggered = False
                    break
            if not all_conditions_triggered:  # clean up if rule not triggered
                self.trigger_column_families = None
                self.trigger_entities = None
            return all_conditions_triggered

    def __repr__(self):
        # Append conditions
        rule_string = "Rule: " + self.name + " has conditions:: "
        is_first = True
        for cond in self.conditions:
            if is_first:
                rule_string += cond
                is_first = False
            else:
                rule_string += (" AND " + cond)
        # Append suggestions
        rule_string += "\nsuggestions:: "
        is_first = True
        for sugg in self.suggestions:
            if is_first:
                rule_string += sugg
                is_first = False
            else:
                rule_string += (", " + sugg)
        if self.trigger_entities:
            rule_string += (', entities:: ' + str(self.trigger_entities))
        if self.trigger_column_families:
            rule_string += (', col_fam:: ' + str(self.trigger_column_families))
        # Return constructed string
        return rule_string


class Suggestion(Section):
    class Action(Enum):
        set = 1
        increase = 2
        decrease = 3

    def __init__(self, name):
        super().__init__(name)
        self.option = None
        self.action = None
        self.suggested_values = None
        self.description = None

    def set_parameter(self, key, value):
        if key == 'option':
            # Note:
            # case 1: 'option' is supported by Rocksdb OPTIONS file; in this
            # case the option belongs to one of the sections in the config
            # file and it's name is prefixed by "<section_type>."
            # case 2: 'option' is not supported by Rocksdb OPTIONS file; the
            # option is not expected to have the character '.' in its name
            self.option = value
        elif key == 'action':
            if self.option and not value:
                raise ValueError(self.name + ': provide action for option')
            self.action = self.Action[value]
        elif key == 'suggested_values':
            if isinstance(value, str):
                self.suggested_values = [value]
            else:
                self.suggested_values = value
        elif key == 'description':
            self.description = value

    def perform_checks(self):
        if not self.description:
            if not self.option:
                raise ValueError(self.name + ': provide option or description')
            if not self.action:
                raise ValueError(self.name + ': provide action for option')
            if self.action is self.Action.set and not self.suggested_values:
                raise ValueError(
                    self.name + ': provide suggested value for option'
                )

    def __repr__(self):
        sugg_string = "Suggestion: " + self.name
        if self.description:
            sugg_string += (' description : ' + self.description)
        else:
            sugg_string += (
                ' option : ' + self.option + ' action : ' + self.action.name
            )
            if self.suggested_values:
                sugg_string += (
                    ' suggested_values : ' + str(self.suggested_values)
                )
        return sugg_string


class Condition(Section):
    def __init__(self, name):
        super().__init__(name)
        self.data_source = None
        self.trigger = None

    def perform_checks(self):
        if not self.data_source:
            raise ValueError(self.name + ': condition not tied to data source')

    def set_data_source(self, data_source):
        self.data_source = data_source

    def get_data_source(self):
        return self.data_source

    def reset_trigger(self):
        self.trigger = None

    def set_trigger(self, condition_trigger):
        self.trigger = condition_trigger

    def get_trigger(self):
        return self.trigger

    def is_triggered(self):
        if self.trigger:
            return True
        return False

    def set_parameter(self, key, value):
        # must be defined by the subclass
        raise NotImplementedError(self.name + ': provide source for condition')


class LogCondition(Condition):
    @classmethod
    def create(cls, base_condition):
        base_condition.set_data_source(DataSource.Type['LOG'])
        base_condition.__class__ = cls
        return base_condition

    def set_parameter(self, key, value):
        if key == 'regex':
            self.regex = value

    def perform_checks(self):
        super().perform_checks()
        if not self.regex:
            raise ValueError(self.name + ': provide regex for log condition')

    def __repr__(self):
        log_cond_str = "LogCondition: " + self.name
        log_cond_str += (" regex: " + self.regex)
        # if self.trigger:
        #     log_cond_str += (" trigger: " + str(self.trigger))
        return log_cond_str


class OptionCondition(Condition):
    @classmethod
    def create(cls, base_condition):
        base_condition.set_data_source(DataSource.Type['DB_OPTIONS'])
        base_condition.__class__ = cls
        return base_condition

    def set_parameter(self, key, value):
        if key == 'options':
            if isinstance(value, str):
                self.options = [value]
            else:
                self.options = value
        elif key == 'evaluate':
            self.eval_expr = value

    def perform_checks(self):
        super().perform_checks()
        if not self.options:
            raise ValueError(self.name + ': options missing in condition')
        if not self.eval_expr:
            raise ValueError(self.name + ': expression missing in condition')

    def __repr__(self):
        opt_cond_str = "OptionCondition: " + self.name
        opt_cond_str += (" options: " + str(self.options))
        opt_cond_str += (" expression: " + self.eval_expr)
        if self.trigger:
            opt_cond_str += (" trigger: " + str(self.trigger))
        return opt_cond_str


class TimeSeriesCondition(Condition):
    @classmethod
    def create(cls, base_condition):
        base_condition.set_data_source(DataSource.Type['TIME_SERIES'])
        base_condition.__class__ = cls
        return base_condition

    def set_parameter(self, key, value):
        if key == 'keys':
            if isinstance(value, str):
                self.keys = [value]
            else:
                self.keys = value
        elif key == 'behavior':
            self.behavior = TimeSeriesData.Behavior[value]
        elif key == 'rate_threshold':
            self.rate_threshold = float(value)
        elif key == 'window_sec':
            self.window_sec = int(value)
        elif key == 'evaluate':
            self.expression = value
        elif key == 'aggregation_op':
            self.aggregation_op = TimeSeriesData.AggregationOperator[value]

    def perform_checks(self):
        if not self.keys:
            raise ValueError(self.name + ': specify timeseries key')
        if not self.behavior:
            raise ValueError(self.name + ': specify triggering behavior')
        if self.behavior is TimeSeriesData.Behavior.bursty:
            if not self.rate_threshold:
                raise ValueError(self.name + ': specify rate burst threshold')
            if not self.window_sec:
                self.window_sec = 300  # default window length is 5 minutes
            if len(self.keys) > 1:
                raise ValueError(self.name + ': specify only one key')
        elif self.behavior is TimeSeriesData.Behavior.evaluate_expression:
            if not (self.expression):
                raise ValueError(self.name + ': specify evaluation expression')
        else:
            raise ValueError(self.name + ': trigger behavior not supported')

    def __repr__(self):
        ts_cond_str = "TimeSeriesCondition: " + self.name
        ts_cond_str += (" statistics: " + str(self.keys))
        ts_cond_str += (" behavior: " + self.behavior.name)
        if self.behavior is TimeSeriesData.Behavior.bursty:
            ts_cond_str += (" rate_threshold: " + str(self.rate_threshold))
            ts_cond_str += (" window_sec: " + str(self.window_sec))
        if self.behavior is TimeSeriesData.Behavior.evaluate_expression:
            ts_cond_str += (" expression: " + self.expression)
            if hasattr(self, 'aggregation_op'):
                ts_cond_str += (" aggregation_op: " + self.aggregation_op.name)
        if self.trigger:
            ts_cond_str += (" trigger: " + str(self.trigger))
        return ts_cond_str


class RulesSpec:
    def __init__(self, rules_path):
        self.file_path = rules_path

    def initialise_fields(self):
        self.rules_dict = {}
        self.conditions_dict = {}
        self.suggestions_dict = {}

    def perform_section_checks(self):
        for rule in self.rules_dict.values():
            rule.perform_checks()
        for cond in self.conditions_dict.values():
            cond.perform_checks()
        for sugg in self.suggestions_dict.values():
            sugg.perform_checks()

    def load_rules_from_spec(self):
        self.initialise_fields()
        with open(self.file_path, 'r') as db_rules:
            curr_section = None
            for line in db_rules:
                line = IniParser.remove_trailing_comment(line)
                if not line:
                    continue
                element = IniParser.get_element(line)
                if element is IniParser.Element.comment:
                    continue
                elif element is not IniParser.Element.key_val:
                    curr_section = element  # it's a new IniParser header
                    section_name = IniParser.get_section_name(line)
                    if element is IniParser.Element.rule:
                        new_rule = Rule(section_name)
                        self.rules_dict[section_name] = new_rule
                    elif element is IniParser.Element.cond:
                        new_cond = Condition(section_name)
                        self.conditions_dict[section_name] = new_cond
                    elif element is IniParser.Element.sugg:
                        new_suggestion = Suggestion(section_name)
                        self.suggestions_dict[section_name] = new_suggestion
                elif element is IniParser.Element.key_val:
                    key, value = IniParser.get_key_value_pair(line)
                    if curr_section is IniParser.Element.rule:
                        new_rule.set_parameter(key, value)
                    elif curr_section is IniParser.Element.cond:
                        if key == 'source':
                            if value == 'LOG':
                                new_cond = LogCondition.create(new_cond)
                            elif value == 'OPTIONS':
                                new_cond = OptionCondition.create(new_cond)
                            elif value == 'TIME_SERIES':
                                new_cond = TimeSeriesCondition.create(new_cond)
                        else:
                            new_cond.set_parameter(key, value)
                    elif curr_section is IniParser.Element.sugg:
                        new_suggestion.set_parameter(key, value)

    def get_rules_dict(self):
        return self.rules_dict

    def get_conditions_dict(self):
        return self.conditions_dict

    def get_suggestions_dict(self):
        return self.suggestions_dict

    def get_triggered_rules(self, data_sources, column_families):
        self.trigger_conditions(data_sources)
        triggered_rules = []
        for rule in self.rules_dict.values():
            if rule.is_triggered(self.conditions_dict, column_families):
                triggered_rules.append(rule)
        return triggered_rules

    def trigger_conditions(self, data_sources):
        for source_type in data_sources:
            cond_subset = [
                cond
                for cond in self.conditions_dict.values()
                if cond.get_data_source() is source_type
            ]
            if not cond_subset:
                continue
            for source in data_sources[source_type]:
                source.check_and_trigger_conditions(cond_subset)

    def print_rules(self, rules):
        for rule in rules:
            print('\nRule: ' + rule.name)
            for cond_name in rule.conditions:
                print(repr(self.conditions_dict[cond_name]))
            for sugg_name in rule.suggestions:
                print(repr(self.suggestions_dict[sugg_name]))
            if rule.trigger_entities:
                print('scope: entities:')
                print(rule.trigger_entities)
            if rule.trigger_column_families:
                print('scope: col_fam:')
                print(rule.trigger_column_families)



"""Module contains decorators for typer cli commands."""
import json
import logging
from functools import wraps

import typer

from cmapi_server.exceptions import CMAPIBasicError


def handle_output(func):
    """Decorator for handling output errors and add result to log file."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger('mcs_cli')
        return_code = 1
        try:
            result = func(*args, **kwargs)
            # Print strings as-is for human-readable output; JSON otherwise
            if isinstance(result, str):
                logging.debug(f'Command returned: {result}')
            else:
                typer.echo(json.dumps(result, indent=2))
            logger.debug(f'Command returned: {result}')
            return_code = 0
        except CMAPIBasicError as err:
            typer.echo(err.message, err=True)
            logger.error('Error during command execution', exc_info=True)
        except typer.BadParameter as err:
            logger.error('Bad command line parameter.')
            raise err
        except typer.Exit as err:  # if some command used typer.Exit
            #TODO: think about universal protocol to return json data and
            #      plain text results.
            return_code = err.exit_code
        except Exception:
            logger.error(
                'Undefined error during command execution',
                exc_info=True
            )
            typer.echo('Unknown error, check the log file.', err=True)

        raise typer.Exit(return_code)
    return wrapper



import logging
import os
from typing import List, Optional, Tuple

from mr_kot import Runner, Status, any_of, check, check_all, fact, parametrize
from mr_kot_fs_validators import Exists, GroupIs, IsDir, IsExecutable, IsReadable, IsWritable, OwnerIs

from cmapi_server import helpers
from cmapi_server.constants import CMAPI_CONF_PATH, MCS_DATA_PATH
from mcs_node_control.models.node_config import NodeConfig

logger = logging.getLogger(__name__)


def resolve_symlinks_in_path(base_path: str, subdirs: List[str]) -> List[str]:
    """Build required directories list, resolving symlinks at each level.

    If base_path or any intermediate directory is a symlink, resolve it
    before appending subsequent path components. This ensures we check
    the actual target directories rather than potentially broken paths.

    :param base_path: The root path (e.g., MCS_DATA_PATH).
    :param subdirs: List of subdirectory components to append sequentially.
    :return: List of resolved directory paths to check.
    """
    result: List[str] = []
    current: str = base_path

    # Resolve base path if it's a symlink
    if os.path.islink(current):
        current = os.path.realpath(current)
        logger.debug('Base path %s is a symlink, resolved to %s', base_path, current)

    result.append(current)

    for subdir in subdirs:
        current = os.path.join(current, subdir)
        # Resolve if this path component is a symlink
        if os.path.islink(current):
            resolved: str = os.path.realpath(current)
            logger.debug('Path %s is a symlink, resolved to %s', current, resolved)
            current = resolved
        result.append(current)

    return result


def is_invariant_checks_enabled() -> bool:
    """Check if invariant checks are enabled in CMAPI config.

    :return: True if invariant checks are enabled (default), False otherwise.
    """
    cfg_parser = helpers.get_config_parser(CMAPI_CONF_PATH)
    return cfg_parser.getboolean('application', 'invariant_checks', fallback=True)


def run_invariant_checks() -> Optional[str]:
    """Run invariant checks, log results, and return a formatted string with problems, if any.

    If invariant checks are disabled in CMAPI config file (invariant_checks = false),
    this function returns None without running any checks.

    :return: Formatted string with problems if checks fail, None otherwise.
    """
    if not is_invariant_checks_enabled():
        logger.info('Invariant checks are turned OFF in CMAPI config file.')
        return None

    logger.info('Starting invariant checks')
    runner = Runner()
    result = runner.run()
    problems = result.problems()

    # Log each fail/error/warning for diagnostics
    diag: str = ''
    for problem in problems:
        fn = logger.warning if problem.status == Status.WARN else logger.error
        fn(
            'Invariant check with id=%s produced %s: %r',
            problem.id, problem.status, problem.evidence,
        )
        diag += f'{problem.id}: {problem.evidence}\n'

    logger.info(
        'Stats: overall=%s counts=%s',
        result.overall.value,
        {k.value: v for k, v in result.counts.items() if v != 0}
    )
    if result.overall in (Status.FAIL, Status.ERROR):
        logger.error('Invariant checks failed')
        return diag
    else:
        logger.info('Invariant checks passed')
        return None


### Facts
@fact
def storage_type() -> str:
    """Provides storage type: shared_fs or s3."""
    return 's3' if NodeConfig().s3_enabled() else 'shared_fs'

@fact
def is_shared_fs(storage_type: str) -> bool:
    return storage_type == 'shared_fs'

@fact
def dispatcher_name() -> str:
    """Provides environment dispatcher name: systemd or container"""
    cfg = helpers.get_config_parser()
    name, _ = helpers.get_dispatcher_name_and_path(cfg)
    return name

@fact
def is_systemd_disp(dispatcher_name: str) -> bool:
    return dispatcher_name == 'systemd'


### Checks
# Build required directories list, resolving symlinks at each level.
# This handles cases where MCS_DATA_PATH or 'data1' is a symlink.
REQUIRED_LOCAL_DIRS = resolve_symlinks_in_path(
    MCS_DATA_PATH,
    ['data1', 'systemFiles', 'dbrm']
)

@check(selector='is_shared_fs')
@parametrize('dir', values=REQUIRED_LOCAL_DIRS, fail_fast=True)
def required_dirs_perms(dir: str) -> Tuple[Status, str]:
    status, ev = check_all(
        dir,
        Exists(),
        IsDir(),
        IsReadable(),
        IsWritable(),
        IsExecutable(),
    )
    return (status, ev)

@check(selector='is_shared_fs, is_systemd_disp')
@parametrize('dir', values=REQUIRED_LOCAL_DIRS, fail_fast=True)
def required_dirs_ownership(dir: str) -> Tuple[Status, str]:
    # Check ownership only when not in containers
    status, ev = check_all(
        dir,
        Exists(),
        # The correct owner is mysql, but i've seen mariadb as owner of the mountpoint,
        #   so we allow both
        any_of(OwnerIs('mysql'), OwnerIs('mariadb')),
        any_of(GroupIs('mysql'), GroupIs('mariadb')),
    )
    return (status, ev)



import logging
import os
from functools import wraps
from typing import Optional, Tuple

logger = logging.getLogger("tracing")

def swallow_exceptions(method):
    """Decorator that logs exceptions and prevents them from propagating up."""

    @wraps(method)
    def _wrapper(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except Exception:
            logger.exception("%s failed", getattr(method, "__qualname__", repr(method)))
            return None

    return _wrapper

def rand_16_hex() -> str:
    """Return 16 random bytes as a 32-char hex string (trace_id size)."""
    return os.urandom(16).hex()


def rand_8_hex() -> str:
    """Return 8 random bytes as a 16-char hex string (span_id size)."""
    return os.urandom(8).hex()


def format_traceparent(trace_id: str, span_id: str, flags: str = "01") -> str:
    """Build a W3C traceparent header (version 00)."""
    return f"00-{trace_id}-{span_id}-{flags}"


def parse_traceparent(header: str) -> Optional[Tuple[str, str, str]]:
    """Parse W3C traceparent and return (trace_id, span_id, flags) or None."""
    try:
        parts = header.strip().split("-")
        if len(parts) != 4 or parts[0] != "00":
            logger.error("Invalid traceparent: %s", header)
            return None
        trace_id, span_id, flags = parts[1], parts[2], parts[3]
        if len(trace_id) != 32 or len(span_id) != 16 or len(flags) != 2:
            return None
        # W3C: all zero trace_id/span_id are invalid
        if set(trace_id) == {"0"} or set(span_id) == {"0"}:
            return None
        return trace_id, span_id, flags
    except Exception:
        logger.exception("Failed to parse traceparent: %s", header)
        return None



"""Async sibling of TracedSession."""
from typing import Any

import logging
import time
import aiohttp

from tracing.tracer import get_tracer

# Limit for raw JSON string preview (in characters)
_PREVIEW_MAX_CHARS = 512

logger = logging.getLogger("tracer")


class TracedAsyncSession(aiohttp.ClientSession):
    async def _request(
        self, method: str, str_or_url: Any, *args: Any, **kwargs: Any
    ) -> aiohttp.ClientResponse:
        tracer = get_tracer()

        headers = kwargs.get("headers") or {}
        if headers is None:
            headers = {}
        kwargs["headers"] = headers

        url_text = str(str_or_url)
        span_name = f"HTTP {method} {url_text}"
        with tracer.start_as_current_span(span_name, kind="CLIENT") as span:
            span.set_attribute("http.method", method)
            span.set_attribute("http.url", url_text)
            span.set_attribute("request_is_sync", False)
            tracer.inject_outbound_headers(headers)
            try:
                response = await super()._request(method, str_or_url, *args, **kwargs)
            except Exception as exc:
                span.set_status("ERROR", str(exc))
                raise
            else:
                span.set_attribute("http.status_code", response.status)
                await _record_outbound_json_preview(response, span)
                return response
            finally:
                duration_ms = (time.time_ns() - span.start_ns) / 1_000_000.0
                span.set_attribute("request_duration_ms", duration_ms)


def create_traced_async_session(**kwargs: Any) -> TracedAsyncSession:
    return TracedAsyncSession(**kwargs)



async def _record_outbound_json_preview(response: aiohttp.ClientResponse, span) -> None:
    """If response is JSON, attach small part of it to span

    We don't use streaming in aiohttp, so reading text is safe here.
    """
    try:
        content_type = str(response.headers.get('Content-Type', '')).lower()
        if 'application/json' not in content_type:
            return
        text = await response.text()
        if text is None:
            text = ""
        span.set_attribute('http.response.body.size', len(text))
        span.set_attribute('http.response.json', text[:_PREVIEW_MAX_CHARS])
    except Exception:
        logger.exception("Could not extract JSON response body")
        return None



import json
from pathlib import Path
import subprocess
from integration_tests.utils import change_directory


def cluster_exists(terraform_dir) -> bool:
    """
    Check if the cluster exists by running terraform output command.
    """
    with change_directory(terraform_dir):
        try:
            outputs = json.loads(subprocess.check_output(
                ["terraform", "output", "-json"],
                text=True,
            ))
            return bool(outputs)
        except subprocess.CalledProcessError as e:
            print(f"Error checking cluster existence: {e}")
            return False


def create_cluster(terraform_dir: Path) -> None:
    """
    Create the cluster by running terraform and ansible
    """
    with change_directory(terraform_dir):
        subprocess.check_call(["terraform", "init"])
        subprocess.check_call(["terraform", "plan"])
        subprocess.check_call(["terraform", "apply", "-auto-approve"])
        subprocess.check_call(["ansible-playbook", "provision.yml"])

        # Run our own provisioning playbook to prepare the cluster for the tests
        our_playbook = Path(__file__).parent / "integration_tests.yml"
        subprocess.check_call(["ansible-playbook", str(our_playbook.absolute())])


def destroy_cluster(terraform_dir: Path) -> None:
    assert cluster_exists(terraform_dir), "Cluster does not exist"
    with change_directory(terraform_dir):
        subprocess.check_call(["terraform", "destroy", "-auto-approve"])



import unittest
from unittest.mock import MagicMock

# pylint: disable=protected-access

from failover.shared_storage_monitor import SharedStorageMonitor
from failover.heartbeat_history import HBHistory


class SharedStorageMonitorTestCase(unittest.TestCase):

    def setUp(self):
        self.hb = HBHistory(tickWindow=10)
        self.monitor = SharedStorageMonitor(check_interval=0, hb_history=self.hb)
        self.monitor._cluster_api_client = MagicMock()

    def test_check_shared_storage_returns_none_when_nodes_unreachable(self):
        self.monitor._cluster_api_client.check_shared_storage.return_value = {
            'shared_storage': False,
            'active_nodes_count': 3,
            'nodes_responses': {
                'node-1': {'success': False, 'error': 'timeout'},
            },
            'nodes_errors': {'node-1': 'timeout'},
        }

        result = self.monitor._check_shared_storage()

        self.assertIsNone(result)
        self.assertEqual({'node-1'}, self.monitor._last_unreachable_nodes)

    def test_check_shared_storage_recovers_after_unreachable_nodes(self):
        # First call reports unreachable nodes
        self.monitor._cluster_api_client.check_shared_storage.return_value = {
            'shared_storage': False,
            'active_nodes_count': 3,
            'nodes_responses': {
                'node-1': {'success': False, 'error': 'timeout'},
            },
            'nodes_errors': {'node-1': 'timeout'},
        }
        first_result = self.monitor._check_shared_storage()
        self.assertIsNone(first_result)
        self.assertEqual({'node-1'}, self.monitor._last_unreachable_nodes)

        # Second call indicates all nodes reachable and shared storage ok
        self.monitor._cluster_api_client.check_shared_storage.return_value = {
            'shared_storage': True,
            'active_nodes_count': 3,
            'nodes_responses': {
                'node-1': {'success': True},
            },
            'nodes_errors': {},
        }
        second_result = self.monitor._check_shared_storage()

        self.assertTrue(second_result)
        self.assertFalse(self.monitor._last_unreachable_nodes)

    def test_skips_nodes_with_recent_good_to_no_transition(self):
        # Instead of relying on environment get_active_nodes, patch the helper
        # to return a deterministic skip list derived from HB signal.
        self.monitor._retrieve_unstable_nodes = MagicMock(return_value=['node-2'])

        # Mock API client to verify payload contains skip_nodes
        self.monitor._cluster_api_client.check_shared_storage.return_value = {
            'shared_storage': True,
            'active_nodes_count': 2,
            'nodes_responses': {
                'node-1': {'success': True},
            },
            'nodes_errors': {},
        }

        _ = self.monitor._check_shared_storage()

        # We expect check_shared_storage to be called with extra payload
        calls = self.monitor._cluster_api_client.check_shared_storage.mock_calls
        # Find latest call with kwargs
        self.assertTrue(calls, 'check_shared_storage was not called')
        last_call = calls[-1]
        # last_call is call(extra={...})
        kwargs = last_call.kwargs
        self.assertIn('extra', kwargs)
        self.assertIn('skip_nodes', kwargs['extra'])
        self.assertEqual(['node-2'], kwargs['extra']['skip_nodes'])


if __name__ == '__main__':
    unittest.main()



"""Cmapi typer application.

Formally this module contains all subcommands for "mcs cmapi" cli command.
"""
import logging
from typing import Optional
from typing_extensions import Annotated

import requests
import typer

from cmapi_server.constants import CMAPI_CONF_PATH
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.helpers import get_config_parser, get_current_key, get_version, build_url
from mcs_cluster_tool.decorators import handle_output


logger = logging.getLogger('mcs_cli')
app = typer.Typer(
    help='Commands related to CMAPI itself.'
)
config_app = typer.Typer(help='Manage CMAPI configuration.')
app.add_typer(config_app, name='config')


@app.command()
@handle_output
def is_ready(
    node: Annotated[
        str,
        typer.Option(
            '--node',
            help=('Which node to check the CMAPI is ready to handle requests.')
        )
    ] = '127.0.0.1'
):
    """Check CMAPI is ready to handle requests."""
    url = f'https://{node}:8640/cmapi/ready'
    try:
        resp = requests.get(url, verify=False, timeout=15)
        resp.raise_for_status()
        r_json = resp.json()
    except requests.exceptions.HTTPError as err:
        if err.response.status_code == 503:
            raise CMAPIBasicError('CMAPI is not ready.') from err
        else:
            raise CMAPIBasicError(
                'Got unexpected HTTP return code '
                f'"{err.response.status_code}" while getting CMAPI ready '
                'state.'
            ) from err
    except Exception as err:
        raise CMAPIBasicError('Got an error getting CMAPI ready state.') from err
    logger.debug('Successfully get CMAPI ready state.')
    return r_json


@config_app.command()
@handle_output
def set(
    sampling_interval_seconds: Annotated[
        Optional[int],
        typer.Option(
            '--sampling-interval-seconds',
            min=1,
            help='Failover sampling interval in seconds.'
        )
    ] = None,
):
    """Set CMAPI configuration on all nodes."""
    if sampling_interval_seconds is None:
        raise typer.BadParameter('Provide at least one option to set.')

    cfg_parser = get_config_parser(CMAPI_CONF_PATH)
    api_key = get_current_key(cfg_parser)
    version = get_version()
    url = build_url(
        base_url='localhost', port=8640,
        path=f'cmapi/{version}/cmapi_config',
        query_params={},
    )
    headers = {'x-api-key': api_key}
    body = {}
    if sampling_interval_seconds is not None:
        body['failover_sampling_interval_seconds'] = sampling_interval_seconds

    try:
        resp = requests.patch(url, verify=False, headers=headers, json=body)
        resp.raise_for_status()
        return resp.json()
    except requests.exceptions.HTTPError as err:
        status = err.response.status_code if err.response is not None else 'unknown'
        raise CMAPIBasicError(
            f'Got HTTP {status} while updating CMAPI config.'
        ) from err
    except Exception as err:
        raise CMAPIBasicError('Failed to update CMAPI config.') from err



from integration_tests.ssh import RemoteHost

def activate_single_node_config(host: RemoteHost):
    """
    Stop MCS services on the host.
    """
    with host.ssh_connection() as conn:
        print(f"{host.name}: Stopping MCS services...")
        conn.sudo("systemctl stop mariadb mcs-workernode@1 mcs-workernode@2 mcs-writeengineserver mcs-storagemanager mcs-ddlproc mcs-dmlproc mcs-primproc mcs-controllernode", hide=True)
        print(f"{host.name}: Activating single node configuration...")
        conn.sudo("cp /usr/share/columnstore/cmapi/cmapi_server/SingleNode.xml /etc/columnstore/Columnstore.xml", hide=True)
        print(f"{host.name}: Truncating log files...")
        truncation_script = """
        for log in crit.log err.log info.log debug.log cmapi_server.log mcs_cli.log; do
            if [ -f /var/log/mariadb/columnstore/$log ]; then
                sudo truncate -s 0 /var/log/mariadb/columnstore/$log
            fi
        done"""
        conn.sudo(truncation_script, hide=True, shell=True)
        print(f"{host.name}: Starting cmapi service...")
        conn.sudo("systemctl start mariadb-columnstore-cmapi", hide=True)

        mcs_output = host.exec_mcs("cluster status")
        assert mcs_output.get("num_nodes") == 0



import os

from cmapi_server.managers.process import MCSProcessManager
from cmapi_server.process_dispatchers.systemd import SystemdDispatcher
from cmapi_server.test.unittest_global import (
    DDL_SERVICE, CONTROLLERNODE_SERVICE, SYSTEMCTL,
    BaseProcessDispatcherCase
)

class SystemdTest(BaseProcessDispatcherCase):

    def test_systemd_status_start(self):
        os.system(f'{SYSTEMCTL} stop {DDL_SERVICE}')
        self.assertFalse(
            SystemdDispatcher.is_service_running(service=DDL_SERVICE, use_sudo=False)
        )
        self.assertTrue(SystemdDispatcher.start(service=DDL_SERVICE, use_sudo=False))

        os.system(f'{SYSTEMCTL} stop {CONTROLLERNODE_SERVICE}')
        self.assertFalse(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )
        result = SystemdDispatcher.start(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        self.assertTrue(result)
        self.assertTrue(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )

    def test_systemd_status_stop(self):
        os.system(f'{SYSTEMCTL} start {CONTROLLERNODE_SERVICE}')
        self.assertTrue(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )
        self.assertTrue(SystemdDispatcher.stop(service=CONTROLLERNODE_SERVICE, use_sudo=False))
        self.assertFalse(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )

    def test_systemd_status_restart(self):
        os.system(f'{SYSTEMCTL} start {CONTROLLERNODE_SERVICE}')
        self.assertTrue(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )
        self.assertTrue(SystemdDispatcher.restart(service=CONTROLLERNODE_SERVICE, use_sudo=False))
        self.assertTrue(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )

        os.system(f'{SYSTEMCTL} stop {CONTROLLERNODE_SERVICE}')
        self.assertFalse(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )
        self.assertTrue(SystemdDispatcher.restart(service=CONTROLLERNODE_SERVICE, use_sudo=False))
        self.assertTrue(
            SystemdDispatcher.is_service_running(service=CONTROLLERNODE_SERVICE, use_sudo=False)
        )


class MCSProcessManagerTest(BaseProcessDispatcherCase):

    def get_systemd_serv_name(self, service_name):
        if service_name == 'mcs-workernode':
            return f'{service_name}@1'
        return service_name

    def test_mcs_process_manager(self):
        MCSProcessManager.detect('systemd', '')
        for prog in MCSProcessManager._get_sorted_progs(is_primary=True, reverse=True).values():
            serv_name = self.get_systemd_serv_name(prog.service_name)
            os.system(f'{SYSTEMCTL} stop {serv_name}')
        self.assertIsNone(MCSProcessManager.start_node(is_primary=True, use_sudo=False))

        for prog in MCSProcessManager.mcs_progs.values():
            serv_name = self.get_systemd_serv_name(prog.service_name)
            if serv_name == 'mcs-storagemanager':
                continue
            self.assertTrue(
                MCSProcessManager.process_dispatcher.is_service_running(
                    service=serv_name, use_sudo=False
                )
            )

        self.assertIsNone(MCSProcessManager.stop_node(is_primary=True, use_sudo=False))
        for prog in MCSProcessManager.mcs_progs.values():
            serv_name = self.get_systemd_serv_name(prog.service_name)
            self.assertFalse(
                MCSProcessManager.process_dispatcher.is_service_running(
                    service=serv_name, use_sudo=False
                )
            )
        self.assertEqual(len(MCSProcessManager.get_running_mcs_procs()), 0)
        self.assertTrue(
            MCSProcessManager.is_node_processes_ok(
                is_primary=True, node_stopped=True
            )
        )

        for prog in MCSProcessManager._get_sorted_progs(is_primary=True).values():
            serv_name = self.get_systemd_serv_name(prog.service_name)
            os.system(f'{SYSTEMCTL} start {serv_name}')

        for prog in MCSProcessManager.mcs_progs.values():
            serv_name = self.get_systemd_serv_name(prog.service_name)
            self.assertTrue(
                MCSProcessManager.process_dispatcher.is_service_running(
                    service=serv_name, use_sudo=False
                )
            )
        self.assertEqual(
            len(MCSProcessManager.get_running_mcs_procs()),
            len(MCSProcessManager.mcs_progs.keys())
        )
        self.assertTrue(
            MCSProcessManager.is_node_processes_ok(
                is_primary=True, node_stopped=False
            )
        )



#!/usr/bin/env python3
"""
Read CMAPI tracing JSONL logs (JsonFormatter) from one or more nodes and render a
PlantUML sequence diagram showing inbound (SERVER) and outbound (CLIENT) requests
between nodes.

Inputs
  - One or more JSONL log files
  - Each input log file is assumed to be from a single node.

Output
  - PlantUML written to the specified output file.

Parsing and correlation
  - Detects span_begin/span_end events from JSONL (msg in {"span_begin","span_end"}).
  - Per file, reads the "my_addresses" record and takes the first non-loopback IP (preferring IPv4)
    as the node's local IP; all spans from that file carry this local IP.
  - Merge spans across files by (trace_id, span_id). Duration and status_code are added when present.

Normalization and aliasing
  - Localhost remap: any 127.0.0.1/::1 seen in that file is replaced with the file's local IP.
  - Remote host normalization: hostnames with embedded IPs (dashed or dotted) are converted to dotted IPs.
  - Aliasing: if an IP is provided via --alias IP=NAME, the diagram uses NAME for that participant.

Diagram semantics
  - SERVER spans: "<remote> --> METHOD /path" render as remote -> local (local is the file's node).
  - CLIENT spans: "HTTP METHOD https://host/path" render as local -> remote.
  - CLIENT spans prefer request_duration_ms; otherwise duration_ms is shown if available.
  - SERVER spans include [status_code] when available.

CLI options
  - --filter-trace TRACE_ID   Only include the specified trace_id.
  - --alias IP=NAME           Map an IP to a friendly name (may be specified multiple times).
  - --list-ips                Print the set of encountered IPs and exit.

Usage examples
  python3 cmapi/scripts/trace_to_plantuml.py json_trace_mcs1 json_trace_mcs2 json_trace_mcs3 --output diagram.puml
  python3 cmapi/scripts/trace_to_plantuml.py json_trace_mcs1 json_trace_mcs2 json_trace_mcs3 --output diagram.puml
      --filter-trace 1f3b2cb2... --alias 172.31.41.127=mcs2 --alias 172.31.45.34=mcs3
"""
from __future__ import annotations

import argparse
import colorsys
import dataclasses
import datetime as dt
import hashlib
import ipaddress
import json
import logging
import os
import re
import sys
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

JSON_TS_FORMATS = [
    "%Y-%m-%d %H:%M:%S,%f",  # default logging formatter
    "%Y-%m-%d %H:%M:%S",     # without millis
]

# Minimum processing time to show activation lifeline (in milliseconds)
ACTIVATION_MIN_MS = 1000

# dashed IP fragment pattern; will convert '172-31-45-34' -> '172.31.45.34'
DASHED_IP_RE = re.compile(r"(?P<dashed>\b\d{1,3}(?:-\d{1,3}){3}\b)")

logger = logging.getLogger("trace_viz")

def _normalize_cmapi_path(path: str) -> str:
    """Remove common CMAPI prefixes like '/cmapi/0.4.0' and '/cmapi/'.
    """
    if not path:
        return "/"
    # normalize double slashes
    while '//' in path:
        path = path.replace('//', '/')
    prefixes = [
        '/cmapi/0.4.0',
        '/cmapi/',
    ]
    for pref in prefixes:
        if path.startswith(pref):
            path = path[len(pref):]
            break
    if not path.startswith('/'):
        path = '/' + path
    return path if path != '//' else '/'

def _hex_color_for_trace(trace_id: str, depth: int) -> str:
    """Derive a stable color from trace_id and vary by depth.

    We map trace_id hash to a base hue and then shift lightness slightly per depth.
    Returns a hex color string like '#a1b2c3'.
    """
    h = hashlib.sha1(trace_id.encode('utf-8')).digest()
    # base hue from first byte, saturation fixed, lightness varies by depth
    base_hue = h[0] / 255.0  # 0..1
    sat = 0.55
    base_lightness = 0.55
    # depth shift: +/- 0.06 per level alternating
    shift = ((depth % 6) - 3) * 0.02  # -0.06..+0.06
    lightness = max(0.35, min(0.75, base_lightness + shift))
    # convert HSL to RGB
    r, g, b = colorsys.hls_to_rgb(base_hue, lightness, sat)
    return f"#{int(r*255):02x}{int(g*255):02x}{int(b*255):02x}"


def _colorized_arrow(arrow: str, color_hex: str) -> str:
    """Insert a PlantUML color into an arrow token.

    Examples:
      '->'   -> '-[#hex]->'
      '-->'  -> '-[#hex]-->'
      '->>'  -> '-[#hex]->>'
      '-->>' -> '-[#hex]-->>'
    """
    if not color_hex:
        return arrow
    if arrow.startswith('--'):
        return f"-[{color_hex}]{arrow}"
    return f"-[{color_hex}]{arrow}"


def _fmt_secs(ms: float) -> str:
    """Format milliseconds as seconds with 2 decimal places."""
    return f"{(ms or 0.0)/1000.0:.2f}s"

@dataclasses.dataclass
class Span:
    trace_id: str
    span_id: str
    parent_span_id: Optional[str]
    kind: str  # SERVER or CLIENT
    name: str
    start_ts: dt.datetime
    duration_ms: Optional[float] = None
    status_code: Optional[int] = None
    request_duration_ms: Optional[float] = None
    local_label: str = "local"
    local_ip: Optional[str] = None
    local_file: Optional[str] = None  # full path to source log file
    # Structured attributes
    http_method: Optional[str] = None
    http_path: Optional[str] = None
    http_url: Optional[str] = None
    client_ip: Optional[str] = None
    request_is_sync: Optional[bool] = None


@dataclasses.dataclass
class Message:
    ts: dt.datetime
    src: str
    dst: str
    label: str
    arrow: str  # '->', '->>', '-->', '-->>'


@dataclasses.dataclass
class Directive:
    ts: dt.datetime
    text: str  # e.g., 'activate mcs1' or 'deactivate mcs1'


def parse_json_ts(ts_str: str) -> Optional[dt.datetime]:
    for fmt in JSON_TS_FORMATS:
        try:
            return dt.datetime.strptime(ts_str, fmt)
        except Exception:
            continue
    return None


def parse_json_span_event(obj: Dict[str, object]) -> Optional[Tuple[str, Dict[str, object]]]:
    """Return (msg, data) if this JSON object is a span_begin/span_end with usable fields."""
    msg = str(obj.get("msg", ""))
    if msg not in {"span_begin", "span_end"}:
        return None
    return msg, obj


def local_label_from_path(path: str) -> str:
    base = os.path.basename(path)
    return base


def collect_known_ips_from_span(sp: "Span", known_ips: set[str]) -> None:
    """Populate known IPs using structured span attributes only."""
    try:
        if sp.kind == "SERVER" and sp.client_ip and is_ip(sp.client_ip):
            known_ips.add(sp.client_ip)
        elif sp.kind == "CLIENT" and sp.http_url:
            try:
                host = urlparse(sp.http_url).hostname
            except Exception:
                host = None
            if host and is_ip(host):
                known_ips.add(host)
    except Exception:
        pass


def normalize_participant(name: str) -> str:
    # PlantUML participant names cannot contain spaces or some punctuation unless quoted.
    # We'll use an alias form: participant "label" as alias, where alias is alnum/underscore.
    # This function returns a safe alias; the label is kept original elsewhere.
    alias = re.sub(r"[^A-Za-z0-9_]", "_", name)
    # Avoid alias starting with a digit
    if alias and alias[0].isdigit():
        alias = f"n_{alias}"
    return alias or "unknown"


# Legacy name-parsing helpers removed; rely on structured attributes.


def is_ip(value: str) -> bool:
    try:
        ipaddress.ip_address(value)
        return True
    except Exception:
        return False


def parse_jsonl_logs(log_paths: List[str]) -> Tuple[List[Span], List[str]]:
    spans_by_key: Dict[Tuple[str, str], Span] = {}
    known_ips: set[str] = set()
    for path in log_paths:
        local_label = local_label_from_path(path)
        # Per-file local IP discovered from a special 'my_addresses' record
        file_local_ip: Optional[str] = None
        if not os.path.exists(path):
            logger.error("Input log file does not exist: %s", path)
            raise FileNotFoundError(path)
        with open(path, "r", encoding="utf-8", errors="ignore") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                try:
                    obj = json.loads(line)
                except Exception as exc:
                    logger.error("Failed to parse JSON line in %s: %s", path, line[:256])
                    raise ValueError(f"JSON parse error in {path}: {exc}")
                # Detect and store local IPs for this file
                if str(obj.get("msg", "")) == "my_addresses":
                    addrs = obj.get("my_addresses")
                    if isinstance(addrs, list):
                        # Pick first non-loopback IPv4 if available, else first non-loopback
                        candidates = [a for a in addrs if isinstance(a, str) and a not in {"127.0.0.1", "::1"}]
                        # Prefer IPv4
                        ipv4 = [a for a in candidates if re.match(r"^\d{1,3}(?:\.\d{1,3}){3}$", a)]
                        chosen = ipv4[0] if ipv4 else (candidates[0] if candidates else None)
                        if chosen:
                            file_local_ip = chosen
                            known_ips.add(chosen)
                        else:
                            logger.error("No non-loopback addresses found in my_addresses for %s: %s", path, addrs)
                            raise ValueError(f"No non-loopback addresses found in my_addresses for {path}: {addrs}")
                    # continue scanning; spans may come later
                    continue
                parsed = parse_json_span_event(obj)
                if not parsed:
                    continue
                msg, data = parsed
                ts_raw = str(data.get("ts", ""))
                ts = parse_json_ts(ts_raw)
                if ts is None:
                    logger.error("Unparseable timestamp '%s' in %s; defaulting to epoch", ts_raw, path)
                    ts = dt.datetime.fromtimestamp(0)
                # Fields from flattened span dict
                name = str(data.get("span_name", data.get("name", "")))
                kind = str(data.get("kind", "")).upper()
                tid = str(data.get("trace_id", ""))
                sid = str(data.get("span_id", ""))
                psid = data.get("parent_span_id")
                key = (tid, sid)

                if msg == "span_begin":
                    if not tid or not sid or not name or not kind:
                        logger.error(
                            "Malformed span_begin in %s (missing fields): tid=%s sid=%s name=%s kind=%s",
                            path, tid, sid, name, kind,
                        )
                        raise ValueError(f"Malformed span_begin in {path} (missing fields): tid={tid} sid={sid} name={name} kind={kind}")
                    if kind not in {"SERVER", "CLIENT"}:
                        logger.error(
                            "Unexpected span kind in span_begin: kind=%s tid=%s sid=%s name=%s (file=%s)",
                            kind, tid, sid, name, path,
                        )
                        raise ValueError(f"Unexpected span kind in span_begin: kind={kind} tid={tid} sid={sid} name={name} (file={path})")
                    sp = Span(
                        trace_id=tid,
                        span_id=sid,
                        parent_span_id=str(psid) if psid else None,
                        kind=kind,
                        name=name,
                        start_ts=ts,
                        local_label=local_label,
                        local_ip=file_local_ip,
                        local_file=path,
                    )
                    # Capture structured fields when present
                    sp.http_method = str(data.get("http.method")) if data.get("http.method") is not None else sp.http_method
                    sp.http_path = str(data.get("http.path")) if data.get("http.path") is not None else sp.http_path
                    sp.http_url = str(data.get("http.url")) if data.get("http.url") is not None else sp.http_url
                    sp.client_ip = str(data.get("client.ip")) if data.get("client.ip") is not None else sp.client_ip
                    # Status code if present
                    if data.get("http.status_code") is not None:
                        try:
                            sp.status_code = int(data.get("http.status_code"))
                        except Exception:
                            pass
                    if data.get("request_is_sync") is not None:
                        try:
                            sp.request_is_sync = bool(data.get("request_is_sync"))
                        except Exception:
                            pass
                    spans_by_key[key] = sp
                    collect_known_ips_from_span(sp, known_ips)
                elif msg == "span_end":
                    sp = spans_by_key.get(key)
                    if sp is None:
                        # Create minimal span if begin was not seen in provided files
                        logger.error(
                            "span_end without matching span_begin (tid=%s sid=%s) in %s; creating synthetic span",
                            tid, sid, path,
                        )
                        sp = Span(
                            trace_id=tid,
                            span_id=sid or f"unknown_{len(spans_by_key)}",
                            parent_span_id=str(psid) if psid else None,
                            kind=kind or "SERVER",
                            name=name,
                            start_ts=ts,
                            local_label=local_label,
                            local_ip=file_local_ip,
                            local_file=path,
                        )
                    spans_by_key[key] = sp
                    collect_known_ips_from_span(sp, known_ips)
                    # Update structured fields from end event (often richer)
                    try:
                        if data.get("http.method") is not None:
                            sp.http_method = str(data.get("http.method"))
                        if data.get("http.path") is not None:
                            sp.http_path = str(data.get("http.path"))
                        if data.get("http.url") is not None:
                            sp.http_url = str(data.get("http.url"))
                        if data.get("client.ip") is not None:
                            sp.client_ip = str(data.get("client.ip"))
                        if data.get("http.status_code") is not None:
                            try:
                                sp.status_code = int(data.get("http.status_code"))
                            except Exception:
                                pass
                        if data.get("request_is_sync") is not None:
                            sp.request_is_sync = bool(data.get("request_is_sync"))
                    except Exception:
                        pass
                    # Duration may be present as duration_ms attribute in JSON; otherwise leave None
                    dur = data.get("duration_ms")
                    try:
                        sp.duration_ms = float(dur) if dur is not None else sp.duration_ms
                        if sp.duration_ms is not None and sp.duration_ms < 0:
                            logger.warning(
                                "Negative duration_ms for span: tid=%s sid=%s duration_ms=%s",
                                sp.trace_id, sp.span_id, sp.duration_ms,
                            )
                    except Exception:
                        pass
                    # Prefer explicit outbound request duration if present
                    try:
                        req_dur = data.get("request_duration_ms")
                        if req_dur is not None:
                            sp.request_duration_ms = float(req_dur)
                            if sp.request_duration_ms < 0:
                                logger.warning(
                                    "Negative request_duration_ms for CLIENT span: tid=%s sid=%s duration_ms=%s",
                                    sp.trace_id, sp.span_id, sp.request_duration_ms,
                                )
                    except Exception:
                        pass
        # Retroactively assign local_ip to any earlier spans from this file that were created
        # before the 'my_addresses' record was seen
        if file_local_ip:
            for sp in spans_by_key.values():
                if sp.local_file == path and sp.local_ip is None:
                    sp.local_ip = file_local_ip
        else:
            logger.error("Did not see my_addresses for file %s; local_ip unknown for its spans", path)
    # Sort spans by start_ts
    spans = sorted(spans_by_key.values(), key=lambda s: s.start_ts)
    # Post-parse validation: check unresolved parent_span_id references
    by_trace: Dict[str, set[str]] = {}
    for sp in spans:
        by_trace.setdefault(sp.trace_id, set()).add(sp.span_id)
    for sp in spans:
        if sp.parent_span_id and sp.parent_span_id not in by_trace.get(sp.trace_id, set()):
            logger.error(
                "Unresolved parent_span_id: tid=%s sid=%s parent_sid=%s (no matching span_begin/end in inputs)",
                sp.trace_id, sp.span_id, sp.parent_span_id,
            )
    return spans, sorted(known_ips)


def normalize_hostname_to_ip(host: str, known_ips: List[str], host_map: Dict[str, str]) -> str:
    # If host contains a dashed IP that matches a known dotted IP, replace with that IP
    # First, use explicit host->ip mappings from the log
    if host in host_map:
        return host_map[host]
    # Prefer explicit matches against our known IPs (dotted or dashed forms)
    for ip in known_ips:
        dashed_ip = ip.replace('.', '-')
        if dashed_ip in host or ip in host:
            return ip
    # If the hostname includes a dashed-quad anywhere, extract and return only the dotted IP
    m = DASHED_IP_RE.search(host)
    if m:
        dashed = m.group("dashed")
        dotted = dashed.replace("-", ".")
        return dotted
    # If the hostname includes a dotted-quad anywhere, return that IP
    m2 = re.search(r"\b(\d{1,3}(?:\.\d{1,3}){3})\b", host)
    if m2:
        return m2.group(1)
    return host


def build_plantuml(spans: List[Span], filter_trace: Optional[str], known_ips: List[str], ip_alias_map: Dict[str, str]) -> str:
    lines: List[str] = []
    logged_replacements: set[Tuple[str, str]] = set()  # (original, replaced)
    lines.append("@startuml")
    # Collect participants from spans
    participants: Dict[str, str] = {}  # alias -> label

    msgs: List[Message] = []
    directives: List[Directive] = []
    unaliased_ips_logged: set[str] = set()

    # Build a per-file preferred local label from alias map by looking at peers in the same file
    file_preferred_label: Dict[str, str] = {}
    file_label_counts: Dict[str, Dict[str, int]] = {}
    for sp in spans:
        fname = sp.local_file or ""
        if not fname:
            continue
        # Consider both SERVER remote and CLIENT host using structured attributes
        if sp.kind == "SERVER" and sp.client_ip:
            norm = normalize_hostname_to_ip(sp.client_ip, known_ips, {})
            if norm in ip_alias_map:
                label = ip_alias_map[norm]
                file_label_counts.setdefault(fname, {}).setdefault(label, 0)
                file_label_counts[fname][label] += 1
        elif sp.kind == "CLIENT" and sp.http_url:
            try:
                host = urlparse(sp.http_url).hostname or ""
            except Exception:
                host = ""
            if host:
                norm = normalize_hostname_to_ip(host, known_ips, {})
                if norm in ip_alias_map:
                    label = ip_alias_map[norm]
                    file_label_counts.setdefault(fname, {}).setdefault(label, 0)
                    file_label_counts[fname][label] += 1
    for fname, counts in file_label_counts.items():
        best_label = max(counts.items(), key=lambda kv: kv[1])[0]
        file_preferred_label[fname] = best_label

    for sp in spans:
        if filter_trace and sp.trace_id != filter_trace:
            continue
        # Compute depth (nesting) via parent chain length within the same trace
        depth = 0
        try:
            seen = set()
            cur = sp.parent_span_id
            while cur:
                if (sp.trace_id, cur) in seen:
                    break
                seen.add((sp.trace_id, cur))
                depth += 1
                parent = next((p for p in spans if p.trace_id == sp.trace_id and p.span_id == cur), None)
                cur = parent.parent_span_id if parent else None
        except Exception:
            depth = 0
        color = _hex_color_for_trace(sp.trace_id, depth)
        trace_tag = f"[{sp.trace_id[:4]}-{sp.span_id[:2]}]"
        time_str = sp.start_ts.strftime("%H:%M:%S")
        # Prefer explicit outbound request duration for CLIENT spans if present
        if sp.kind == "CLIENT" and sp.request_duration_ms is not None:
            dur_str = f" ({_fmt_secs(sp.request_duration_ms)})"
        else:
            dur_str = f" ({_fmt_secs(sp.duration_ms)})" if sp.duration_ms is not None else ""

        if sp.kind == "SERVER":
            # Structured attributes required: client_ip, http_method, http_path
            remote = sp.client_ip
            method = sp.http_method
            path = sp.http_path
            if not (remote and method and path):
                logger.warning(
                    "Skipping SERVER span due to missing structured fields: tid=%s sid=%s name=%s client.ip=%s http.method=%s http.path=%s",
                    sp.trace_id, sp.span_id, sp.name, sp.client_ip, sp.http_method, sp.http_path,
                )
                continue
            label = f"{method} {_normalize_cmapi_path(path)}"
            # Server status code
            code_str = f" [{sp.status_code}]" if sp.status_code is not None else ""
            if sp.status_code is None:
                logger.warning(
                    "Missing http.status_code for SERVER span at render time: tid=%s sid=%s method=%s path=%s",
                    sp.trace_id, sp.span_id, method, path,
                )
            # Replace loopback with file's real IP if available
            if remote in {"127.0.0.1", "::1"} and sp.local_ip:
                orig = remote
                remote = sp.local_ip
                key = (orig, remote)
                if key not in logged_replacements:
                    file_label = os.path.basename(sp.local_file) if sp.local_file else "?"
                    logger.info("Loopback remapped (SERVER): %s -> %s [file=%s]", orig, remote, file_label)
                    logged_replacements.add(key)
            # do not filter self traffic
            # If remote is an IP, keep as-is; else use hostname
            # apply aliasing for IPs
            # Log unaliased IPs
            if is_ip(remote) and remote not in ip_alias_map and remote not in unaliased_ips_logged:
                logger.info("Unaliased IP encountered (SERVER remote): %s. Consider providing --alias %s=NAME", remote, remote)
                unaliased_ips_logged.add(remote)
            remote_label = ip_alias_map.get(remote, remote)
            remote_alias = normalize_participant(remote_label)
            participants.setdefault(remote_alias, remote_label)
            # Prefer local IP alias (e.g., mcs1); otherwise use the local IP string.
            if sp.local_ip:
                if is_ip(sp.local_ip) and sp.local_ip not in ip_alias_map and sp.local_ip not in unaliased_ips_logged:
                    logger.info("Unaliased IP encountered (SERVER local): %s. Consider providing --alias %s=NAME", sp.local_ip, sp.local_ip)
                    unaliased_ips_logged.add(sp.local_ip)
                local_label_effective = ip_alias_map.get(sp.local_ip, sp.local_ip)
            else:
                # As a last resort, use filename label
                local_label_effective = sp.local_label
            local_alias = normalize_participant(local_label_effective)
            participants.setdefault(local_alias, local_label_effective)
            arrow = _colorized_arrow('->', color)  # SERVER handling is synchronous from remote to local
            msgs.append(Message(sp.start_ts, remote_alias, local_alias, f"{trace_tag} [{time_str}] {label}{dur_str}{code_str}", arrow))
            # Add a note for the request with trace tag and start timestamp
            directives.append(Directive(sp.start_ts, f"note over {remote_alias}, {local_alias}: {trace_tag} [{time_str}]"))
            # Activate server processing lifeline only if duration is known and >= threshold
            if sp.duration_ms is not None and sp.duration_ms >= ACTIVATION_MIN_MS:
                directives.append(Directive(sp.start_ts, f"activate {local_alias} {color}"))
                try:
                    srv_end = sp.start_ts + dt.timedelta(milliseconds=sp.duration_ms)
                    directives.append(Directive(srv_end, f"deactivate {local_alias}"))
                except Exception:
                    pass
            elif sp.duration_ms is None:
                # Log that we couldn't determine server processing duration
                logger.warning(
                    "No duration for SERVER span: tid=%s sid=%s method=%s path=%s local=%s",
                    sp.trace_id, sp.span_id, method, path, local_label_effective,
                )
        elif sp.kind == "CLIENT":
            # Structured attributes required: http_url, http_method
            if not (sp.http_url and sp.http_method):
                logger.warning(
                    "Skipping CLIENT span due to missing structured fields: tid=%s sid=%s name=%s http.url=%s http.method=%s",
                    sp.trace_id, sp.span_id, sp.name, sp.http_url, sp.http_method,
                )
                continue
            # Extract host and path from URL
            try:
                p = urlparse(sp.http_url)
                host = p.hostname or sp.http_url
                path = p.path or "/"
            except Exception:
                logger.warning("Failed to parse URL for CLIENT span (tid=%s sid=%s): %s", sp.trace_id, sp.span_id, sp.http_url)
                host = sp.http_url
                path = sp.http_url
            label = f"{sp.http_method} {_normalize_cmapi_path(path)}"
            # Replace loopback with file's real IP if available
            if host in {"127.0.0.1", "::1"} and sp.local_ip:
                orig = host
                host = sp.local_ip
                key = (orig, host)
                if key not in logged_replacements:
                    file_label = os.path.basename(sp.local_file) if sp.local_file else "?"
                    logger.info("Loopback remapped (CLIENT): %s -> %s [file=%s]", orig, host, file_label)
                    logged_replacements.add(key)
            # Normalize hostnames that embed an IP in dashed form to dotted IP when recognized
            new_host = normalize_hostname_to_ip(host, known_ips, {})
            if new_host != host:
                key = (host, new_host)
                if key not in logged_replacements:
                    file_label = os.path.basename(sp.local_file) if sp.local_file else "?"
                    logger.info("Host normalized: %s -> %s [file=%s]", host, new_host, file_label)
                    logged_replacements.add(key)
            host = new_host
            # do not filter self traffic
            # If host resolved to IP, allow user alias to apply
            if is_ip(host) and host not in ip_alias_map and host not in unaliased_ips_logged:
                logger.info("Unaliased IP encountered (CLIENT host): %s. Consider providing --alias %s=NAME", host, host)
                unaliased_ips_logged.add(host)
            remote_label = ip_alias_map.get(host, host)
            remote_alias = normalize_participant(remote_label)
            participants.setdefault(remote_alias, remote_label)
            # Prefer local IP alias (e.g., mcs1); otherwise use the local IP string.
            if sp.local_ip:
                if is_ip(sp.local_ip) and sp.local_ip not in ip_alias_map and sp.local_ip not in unaliased_ips_logged:
                    logger.info("Unaliased IP encountered (CLIENT local): %s. Consider providing --alias %s=NAME", sp.local_ip, sp.local_ip)
                    unaliased_ips_logged.add(sp.local_ip)
                local_label_effective = ip_alias_map.get(sp.local_ip, sp.local_ip)
            else:
                # As a last resort, use filename label
                local_label_effective = sp.local_label
            local_alias = normalize_participant(local_label_effective)
            participants.setdefault(local_alias, local_label_effective)
            # Use async arrow for async requests
            is_sync = sp.request_is_sync
            arrow = _colorized_arrow('->' if (is_sync is True) else '->>' if (is_sync is False) else '->', color)
            msgs.append(Message(sp.start_ts, local_alias, remote_alias, f"{trace_tag} [{time_str}] {label}{dur_str}", arrow))
            # Add a response arrow back at end time if we know duration
            end_ts: Optional[dt.datetime] = None
            if sp.request_duration_ms is not None:
                try:
                    end_ts = sp.start_ts + dt.timedelta(milliseconds=sp.request_duration_ms)
                except Exception:
                    end_ts = None
            elif sp.duration_ms is not None:
                try:
                    end_ts = sp.start_ts + dt.timedelta(milliseconds=sp.duration_ms)
                except Exception:
                    end_ts = None
            if end_ts is not None:
                end_time_str = end_ts.strftime("%H:%M:%S")
                resp_arrow = '-->' if (is_sync is True or is_sync is None) else '-->>'
                # Build details once
                dur_val = None
                if sp.request_duration_ms is not None:
                    dur_val = sp.request_duration_ms
                elif sp.duration_ms is not None:
                    dur_val = sp.duration_ms
                same_second = (end_time_str == time_str)
                # Response label: always include method/path; timestamp first
                resp_label_parts = ["Response", f"[{end_time_str}]", label]
                if sp.status_code is not None:
                    code_token = str(sp.status_code)
                    if sp.status_code != 200:
                        code_token = f"<color:red>{code_token}</color>"
                    resp_label_parts.append(f"[{code_token}]")
                else:
                    logger.error(
                        "Missing status_code for CLIENT response: tid=%s sid=%s url=%s",
                        sp.trace_id, sp.span_id, sp.http_url,
                    )
                resp_label = ' '.join(resp_label_parts)
                msgs.append(Message(end_ts, remote_alias, local_alias, f"{trace_tag} {resp_label}", _colorized_arrow(resp_arrow, color)))
                if same_second:
                    # Single combined note at the (end) time with tag and duration only
                    parts = [f"{trace_tag}"]
                    if dur_val is not None:
                        parts.append(f"dur={_fmt_secs(dur_val)}")
                    note_text = f"note over {local_alias}, {remote_alias}: " + ", ".join(parts)
                    directives.append(Directive(end_ts, note_text))
                else:
                    # Two separate notes: request note at start, response note at end
                    directives.append(Directive(sp.start_ts, f"note over {local_alias}, {remote_alias}: {trace_tag} [{time_str}]"))
                    parts = [f"{trace_tag}"]
                    if dur_val is not None:
                        parts.append(f"dur={_fmt_secs(dur_val)}")
                    note_text = f"note over {remote_alias}, {local_alias}: " + ", ".join(parts)
                    directives.append(Directive(end_ts, note_text))
            else:
                # No end time available; emit only the request note at start
                directives.append(Directive(sp.start_ts, f"note over {local_alias}, {remote_alias}: {trace_tag} [{time_str}]"))
                logger.info(
                    "No duration for CLIENT request: tid=%s sid=%s method=%s url=%s",
                    sp.trace_id, sp.span_id, sp.http_method, sp.http_url,
                )

    # Emit participants sorted by label (case-insensitive)
    items = list(participants.items())  # (alias, label)
    items.sort(key=lambda x: x[1].lower())

    def emit_part(alias: str, label: str):
        if alias == label and alias.isidentifier():
            lines.append(f"participant {alias}")
        else:
            lines.append(f"participant \"{label}\" as {alias}")

    for alias, label in items:
        emit_part(alias, label)

    # Sort and emit messages and directives merged by time
    msgs.sort(key=lambda m: m.ts)
    directives.sort(key=lambda d: d.ts)
    i = j = 0
    while i < len(msgs) or j < len(directives):
        if j >= len(directives) or (i < len(msgs) and msgs[i].ts <= directives[j].ts):
            m = msgs[i]
            lines.append(f"{m.src} {m.arrow} {m.dst} : {m.label}")
            i += 1
        else:
            d = directives[j]
            lines.append(d.text)
            j += 1

    lines.append("@enduml")
    return "\n".join(lines) + "\n"


def main(argv: Optional[List[str]] = None) -> int:
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
    ap = argparse.ArgumentParser(description="Convert CMAPI tracing JSONL logs to PlantUML sequence diagram")
    ap.add_argument("paths", nargs="+", help="One or more JSONL log files.")
    ap.add_argument("--filter-trace", help="Only include spans for the given trace_id (tid)")
    ap.add_argument("--alias", action="append", default=[], metavar="IP=NAME", help="Assign a label NAME to IP. May be used multiple times.")
    ap.add_argument("--list-ips", action="store_true", help="Print only the set of IP addresses encountered and exit")
    ap.add_argument("--output", default="diagram.puml", help="Output PlantUML path (default: diagram.puml)")
    args = ap.parse_args(argv)

    if args.list_ips:
        logs = args.paths
        if not logs:
            ap.error("At least one log path is required with --list-ips")
        spans, known_ips = parse_jsonl_logs(logs)
    else:
        # Determine logs and output path
        logs = args.paths
        output_path = args.output
        if not logs:
            ap.error("Provide at least one JSONL log path")
        spans, known_ips = parse_jsonl_logs(logs)

    # Build alias map from args.alias entries (used by both list-ips and diagram output)
    ip_alias_map: Dict[str, str] = {}
    for item in args.alias:
        if "=" in item:
            ip, name = item.split("=", 1)
            ip = ip.strip()
            name = name.strip()
            if ip:
                ip_alias_map[ip] = name
        else:
            logger.error("Invalid alias format: %s", item)
            return 2

    if args.list_ips:
        to_print = sorted(set(known_ips))
        for ip in to_print:
            parts = [ip]
            if ip in ip_alias_map:
                parts.append(f"alias={ip_alias_map[ip]}")
            print(" ".join(parts))
        return 0


    puml = build_plantuml(spans, args.filter_trace, known_ips, ip_alias_map)

    if args.list_ips:
        # Already handled above, but keep structure consistent
        return 0

    # Write to output file
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(puml)
    logger.info("Diagram written to %s", output_path)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
try:
    from builtins import object
    from builtins import str
except ImportError:
    from __builtin__ import object
    from __builtin__ import str
import targets_cfg

def pretty_list(lst, indent=8):
    if lst is None or len(lst) == 0:
        return ""

    if len(lst) == 1:
        return "\"%s\"" % lst[0]

    separator = "\",\n%s\"" % (" " * indent)
    res = separator.join(sorted(lst))
    res = "\n" + (" " * indent) + "\"" + res + "\",\n" + (" " * (indent - 4))
    return res


class TARGETSBuilder(object):
    def __init__(self, path, extra_argv):
        self.path = path
        self.targets_file = open(path, 'wb')
        header = targets_cfg.rocksdb_target_header_template.format(
            extra_argv=extra_argv)
        self.targets_file.write(header.encode("utf-8"))
        self.total_lib = 0
        self.total_bin = 0
        self.total_test = 0
        self.tests_cfg = ""

    def __del__(self):
        self.targets_file.close()

    def add_library(self, name, srcs, deps=None, headers=None,
                    extra_external_deps="", link_whole=False):
        headers_attr_prefix = ""
        if headers is None:
            headers_attr_prefix = "auto_"
            headers = "AutoHeaders.RECURSIVE_GLOB"
        else:
            headers = "[" + pretty_list(headers) + "]"
        self.targets_file.write(targets_cfg.library_template.format(
            name=name,
            srcs=pretty_list(srcs),
            headers_attr_prefix=headers_attr_prefix,
            headers=headers,
            deps=pretty_list(deps),
            extra_external_deps=extra_external_deps,
            link_whole=link_whole).encode("utf-8"))
        self.total_lib = self.total_lib + 1

    def add_rocksdb_library(self, name, srcs, headers=None):
        headers_attr_prefix = ""
        if headers is None:
            headers_attr_prefix = "auto_"
            headers = "AutoHeaders.RECURSIVE_GLOB"
        else:
            headers = "[" + pretty_list(headers) + "]"
        self.targets_file.write(targets_cfg.rocksdb_library_template.format(
            name=name,
            srcs=pretty_list(srcs),
            headers_attr_prefix=headers_attr_prefix,
            headers=headers).encode("utf-8"))
        self.total_lib = self.total_lib + 1

    def add_binary(self, name, srcs, deps=None):
        self.targets_file.write(targets_cfg.binary_template.format(
            name=name,
            srcs=pretty_list(srcs),
            deps=pretty_list(deps)).encode("utf-8"))
        self.total_bin = self.total_bin + 1

    def add_c_test(self):
        self.targets_file.write(b"""
cpp_binary(
    name = "c_test_bin",
    srcs = ["db/c_test.c"],
    arch_preprocessor_flags = ROCKSDB_ARCH_PREPROCESSOR_FLAGS,
    compiler_flags = ROCKSDB_COMPILER_FLAGS,
    include_paths = ROCKSDB_INCLUDE_PATHS,
    os_preprocessor_flags = ROCKSDB_OS_PREPROCESSOR_FLAGS,
    preprocessor_flags = ROCKSDB_PREPROCESSOR_FLAGS,
    deps = [":rocksdb_test_lib"],
) if not is_opt_mode else None

custom_unittest(
    name = "c_test",
    command = [
        native.package_name() + "/buckifier/rocks_test_runner.sh",
        "$(location :{})".format("c_test_bin"),
    ],
    type = "simple",
) if not is_opt_mode else None
""")

    def register_test(self,
                      test_name,
                      src,
                      is_parallel,
                      extra_deps,
                      extra_compiler_flags):
        exec_mode = "serial"
        if is_parallel:
            exec_mode = "parallel"
        self.tests_cfg += targets_cfg.test_cfg_template % (
            test_name,
            str(src),
            str(exec_mode),
            extra_deps,
            extra_compiler_flags)

        self.total_test = self.total_test + 1

    def flush_tests(self):
        self.targets_file.write(targets_cfg.unittests_template.format(tests=self.tests_cfg).encode("utf-8"))
        self.tests_cfg = ""



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

from abc import ABC, abstractmethod
import re


class BenchmarkRunner(ABC):
    @staticmethod
    @abstractmethod
    def is_metric_better(new_metric, old_metric):
        pass

    @abstractmethod
    def run_experiment(self):
        # should return a list of DataSource objects
        pass

    @staticmethod
    def get_info_log_file_name(log_dir, db_path):
        # Example: DB Path = /dev/shm and OPTIONS file has option
        # db_log_dir=/tmp/rocks/, then the name of the log file will be
        # 'dev_shm_LOG' and its location will be /tmp/rocks. If db_log_dir is
        # not specified in the OPTIONS file, then the location of the log file
        # will be /dev/shm and the name of the file will be 'LOG'
        file_name = ''
        if log_dir:
            # refer GetInfoLogPrefix() in rocksdb/util/filename.cc
            # example db_path: /dev/shm/dbbench
            file_name = db_path[1:]  # to ignore the leading '/' character
            to_be_replaced = re.compile('[^0-9a-zA-Z\-_\.]')
            for character in to_be_replaced.findall(db_path):
                file_name = file_name.replace(character, '_')
            if not file_name.endswith('_'):
                file_name += '_'
        file_name += 'LOG'
        return file_name



"""
CherryPy tool that uses the tracer to start a span for each request.
"""
import socket
import json
import logging
from typing import Dict, Any

import cherrypy

from tracing.utils import swallow_exceptions
from tracing.tracer import get_tracer

logger = logging.getLogger("tracer")

# Limit for raw JSON string preview (in characters)
_PREVIEW_MAX_CHARS = 512

@swallow_exceptions
def _on_request_start() -> None:
    req = cherrypy.request
    tracer = get_tracer()

    headers: Dict[str, str] = dict(req.headers or {})
    tracer.notify_incoming_request(
        headers=headers,
        method=getattr(req, 'method', ''),
        path=getattr(req, 'path_info', '')
    )
    trace_id, parent_span_id = tracer.extract_traceparent(headers)
    tracer.set_incoming_context(trace_id, parent_span_id)

    requester_host = getattr(getattr(req, 'remote', None), 'ip', '')

    method = getattr(req, 'method', 'HTTP')
    path = getattr(req, 'path_info', '/')
    if requester_host:
        span_name = f"{requester_host} --> {method} {path}"
    else:
        span_name = f"{method} {path}"

    attrs = {
        'http.method': getattr(req, 'method', ''),
        'http.path': getattr(req, 'path_info', ''),
        'client.ip': requester_host,
        'instance.hostname': socket.gethostname(),
    }
    attrs.update(_record_incoming_json_preview(req))

    ctx = tracer.start_as_current_span(span_name, kind="SERVER", attributes=attrs)
    span = ctx.__enter__()
    safe_headers = {k: v for k, v in headers.items() if k.lower() not in {'authorization', 'x-api-key'}}
    span.set_attribute('sentry.incoming_headers', safe_headers)
    req._trace_span_ctx = ctx
    req._trace_span = span

    tracer.inject_traceparent(cherrypy.response.headers)

@swallow_exceptions
def _on_request_end() -> None:
    req = cherrypy.request
    try:
        status_str = str(cherrypy.response.status)
        status_code = int(status_str.split()[0])
    except Exception:
        status_code = None

    span = getattr(req, "_trace_span", None)
    if span is not None and status_code is not None:
        span.set_attribute('http.status_code', status_code)

    tracer = get_tracer()
    tracer.notify_request_finished(status_code)

    ctx = getattr(req, "_trace_span_ctx", None)
    if ctx is not None:
        try:
            ctx.__exit__(None, None, None)
        finally:
            req._trace_span_ctx = None
            req._trace_span = None


def register_tracing_tools() -> None:
    cherrypy.tools.trace = cherrypy.Tool("on_start_resource", _on_request_start, priority=70)
    cherrypy.tools.trace_end = cherrypy.Tool("on_end_resource", _on_request_end, priority=80)

@swallow_exceptions
def _record_incoming_json_preview(req) -> dict[str, Any]:
    """If request Content-Type is JSON, attach a compact preview to span.
    Returns dict of span attributes to set
    """
    attrs = {}
    content_type = str(getattr(req, 'headers', {}).get('Content-Type', '') or '')
    attrs['http.request.content_type'] = content_type
    if 'json' not in content_type.lower():
        return attrs

    parsed_json = getattr(req, 'json', None)
    if parsed_json is None:
        return attrs
    normalized = json.dumps(parsed_json, ensure_ascii=False, sort_keys=True)
    if len(normalized) > _PREVIEW_MAX_CHARS:
        normalized = normalized[:_PREVIEW_MAX_CHARS] + '...<truncated>'
    attrs['http.request.json'] = normalized
    attrs['http.request.body.size'] = len(normalized)
    return attrs



import logging
import re
from dataclasses import dataclass
from time import sleep
from typing import Optional, List

from cmapi_server.constants import SHMEM_LOCKS_PATH
from cmapi_server.process_dispatchers.base import BaseDispatcher


@dataclass
class LocksState:
    id: int
    name: Optional[str]
    readers: int = 0
    writers: int = 0
    readers_waiting: int = 0
    writers_waiting: int = 0
    mutex_locked: bool = False

    def __str__(self):
        name = f"({self.name})" if self.name else ""
        return (f"LS {self.id}{name}: {self.readers}r/{self.writers}w" +
                f" {self.readers_waiting}rw/{self.writers_waiting}ww m={int(self.mutex_locked)}")


def parse_locks_detail(cmd_output: str, logger: logging.Logger) -> List[LocksState]:
    """Parse detailed per-lock state from mcs-shmem-locks output.

    Missing or malformed numeric fields are treated as 0. A logger must be provided
    and will be used to emit warnings for malformed lines.
    """
    states: List[LocksState] = []
    current: Optional[LocksState] = None
    current_id = 0

    for raw in cmd_output.splitlines():
        line = raw.strip()
        if not line:
            continue

        if line.endswith('RWLock'):
            # push previous
            if current is not None:
                states.append(current)
            current_id += 1
            name = line[:-len('RWLock')].strip() or None
            current = LocksState(id=current_id, name=name)
            continue

        if current is None:
            continue

        field_specs = [
            (r'^readers\s*=\s*(\d+)$', 'readers'),
            (r'^writers\s*=\s*(\d+)$', 'writers'),
            (r'^readers waiting\s*=\s*(\d+)$', 'readers_waiting'),
            (r'^writers waiting\s*=\s*(\d+)$', 'writers_waiting'),
        ]

        matched = False
        for pattern, attr in field_specs:
            m = re.search(pattern, line)
            if m:
                try:
                    setattr(current, attr, int(m.group(1)))
                except ValueError:
                    logger.warning('Failed to parse %s from line: %s', attr, raw)
                    setattr(current, attr, 0)
                matched = True
                break
        if matched:
            continue

        m = re.search(r'^mutex locked\s*=\s*(\d+)$', line)
        if m:
            try:
                current.mutex_locked = int(m.group(1)) != 0
            except ValueError:
                current.mutex_locked = False
            continue

    # push the last one
    if current is not None:
        states.append(current)

    return states


def release_shmem_locks(logger: logging.Logger, max_iterations: int = 5) -> bool:
    """Attempt to release active shmem locks.

    - Inspect all locks.
    - Unlock writer lock (there can be only one active, but there can be multiple waiting)
    - Unlock each reader lock sequentially
    - Re-check and repeat up to max_iterations.

    Returns True on success (no active readers/writers remain), False otherwise.
    """
    # We adapt attempts/sleep when there are waiting locks to allow promotions
    attempt = 0
    while True:
        attempt += 1
        success, out = BaseDispatcher.exec_command(f'{SHMEM_LOCKS_PATH} --lock-id 0')
        if not success or not out:
            logger.error('Failed to inspect shmem locks during unlock (attempt %d)', attempt)
            return False

        states = parse_locks_detail(out, logger=logger)
        active_total = sum(s.readers + s.writers for s in states)
        waiting_total = sum(s.readers_waiting + s.writers_waiting for s in states)
        if active_total == 0 and waiting_total == 0:
            attempts_descr = "" if not attempt else f" after {attempt} attempts"
            logger.info('No active or waiting shmem locks remain%s', attempts_descr)
            return True

        logger.debug(
            'Lock release attempt %d: active=%d waiting=%d; detailed=%s',
            attempt, active_total, waiting_total, states
        )

        for st in states:
            if st.writers > 0:
                cmd = f'{SHMEM_LOCKS_PATH} -i {st.id} -w -u'
                ok, _ = BaseDispatcher.exec_command(cmd)
                if not ok:
                    logger.warning('Failed to unlock writer for lock-id=%d', st.id)

            if st.readers > 0:
                for _ in range(st.readers):
                    cmd = f'{SHMEM_LOCKS_PATH} -i {st.id} -r -u'
                    ok, _ = BaseDispatcher.exec_command(cmd)
                    if not ok:
                        logger.warning('Failed to unlock a reader for lock-id=%d', st.id)
                        break

        # Wait for state to settle; longer if we have waiting locks
        sleep(2 if waiting_total > 0 else 1)

        # Allow more attempts when there are waiting locks
        effective_max = max_iterations if waiting_total == 0 else max(max_iterations, 15)
        if attempt >= effective_max:
            break

    logger.error('Failed to fully release shmem locks using mcs-shmem-locks after %d attempts (active/waiting remain)', attempt)
    return False


def sum_active_from_states(states: List[LocksState]) -> int:
    return sum(s.readers + s.writers for s in states)


def sum_waiting_from_states(states: List[LocksState]) -> int:
    return sum(s.readers_waiting + s.writers_waiting for s in states)



import logging
import threading
import time
from socket import socket, SOCK_DGRAM
from struct import pack, unpack_from


class HeartBeater:
    port = 9051
    dieMsg = b'die!00'
    areYouThereMsg = b'AYTM'
    yesIAmMsg = b'YIAM'

    def __init__(self, config, history):
        self.config = config
        self.die = False
        self.history = history
        self.sequenceNum = 0
        self.responseThread = None
        self.sock = None
        self.sockMutex = threading.Lock()
        self.logger = logging.getLogger('heartbeater')

    def start(self):
        self.initSockets()
        self.die = False
        self.responseThread = threading.Thread(
            target=self.listenAndRespond, name='HeartBeater'
        )
        self.responseThread.start()

    def stop(self):
        self.die = True
        # break out of the recv loop
        sock = socket(type=SOCK_DGRAM)
        sock.sendto(self.dieMsg, ('localhost', self.port))
        time.sleep(1)
        self.sock.close()
        self.responseThread.join()

    def initSockets(self):
        self.sock = socket(type=SOCK_DGRAM)
        self.sock.bind(('0.0.0.0', self.port))

    def listenAndRespond(self):
        self.logger.info('Starting the heartbeat listener.')
        while not self.die:
            try:
                self._listenAndRespond()
            except Exception:
                self.logger.warning(
                    'Caught an exception while listening and responding.',
                    exc_info=True
                )
                time.sleep(1)
        self.logger.info('Heartbeat listener exiting normally...')

    def _listenAndRespond(self):
        (data, remote) =  self.sock.recvfrom(300)
        if len(data) < 6:
            return
        (msg_type, seq) = unpack_from('4sH', data, 0)
        if msg_type == self.areYouThereMsg:
            self.logger.trace(f'Got "are you there?" from {remote[0]}')
            name = self.config.who_am_I()
            if name is None:
                self.logger.warning(
                    'Heartbeater: got an "are you there?" msg from '
                    f'{remote[0]}, but this node is not in the list of '
                    'desired nodes for the cluster. '
                    'This node needs a config update.'
                )
                return
            bname = name.encode('ascii')
            if len(bname) > 255:
                bname = bname[:255]
            msg = pack(f'4sH{len(bname)}s', self.yesIAmMsg, seq, bname)
            self.send(msg, remote[0])
            self.logger.trace(f'Send "yes I Am" to {remote[0]}')
        elif msg_type == self.yesIAmMsg:
            if len(data) > 6:
                name = data[6:].decode('ascii')
                self.logger.trace(f'Got "yes I am" from {name}')
                self.history.gotHeartbeat(name, seq)

    def send(self, msg, destaddr):
        self.sockMutex.acquire()
        try:
            self.sock.sendto(msg, (destaddr, self.port))
        except Exception:
            self.logger.warning(
                f'Heartbeater.send(): caught error sending msg to {destaddr}',
                exc_info=True
            )
        finally:
            self.sockMutex.release()

    def sendHeartbeats(self):
        nodes = self.config.getDesiredNodes()
        my_name = self.config.who_am_I()
        msg = pack('4sH', self.areYouThereMsg, self.sequenceNum)
        self.sockMutex.acquire()
        for node in nodes:
            if node == my_name:
                continue
            try:
                self.logger.trace(f'Send "are you there" node {node}')
                self.sock.sendto(msg, (node, self.port))
            except Exception as e:
                pass
                # Suppressing these logs.
                # In docker the whole dns entry gets removed when a container
                # goes away.
                # Ends up spamming the logs until the node is removed from
                # the cluster via the rest endpoint, or the node comes back up.
                # self.logger.warning("Heartbeater.sendHeartbeats():
                # caught an exception sending heartbeat to {}: {}".
                # format(node, e))
        self.sockMutex.release()
        self.sequenceNum = (self.sequenceNum + 1) % 65535
        self.history.setCurrentTick(self.sequenceNum)



"""Tracing support for CMAPI

Despite having many files, the idea of this package is simple: MCS is a distributed system,
  and we need to be able to trace requests across the system.
We need to understand:
* how one incoming request caused many others
* how long each request took
* which request each log line corresponds to
etc

The basic high-level mechanism is this:
1. Each incoming request is assigned a trace ID (or it may already have one, see point 2).
2. This trace ID is propagated to all other outbound requests that are caused by this request.
3. Each sub-operation is assigned a span ID. Request ID stays the same, but the span ID changes.
4. Each span can have a parent span ID, which is the span ID of the request that caused this span.
5. So basically, we have a tree of spans, and the trace ID identifies the root of the tree.

TraceID/SpanID/ParentSpanID are added to each log line, so we can identify which request each log line corresponds to.

Trace attributes are passed through the system via request headers, and here it becomes a bit more complicated.
There are two technologies that we use to pass these ids:
1. W3C TraceContext. This is a standard, it has a fixed header and its format.
   The header is called `traceparent`. It encapsulates trace id and span id.
2. Sentry. For historical reasons, it has different headers. And in our setup it is optional.
   But Sentry is very useful, we also use it to monitor the errors, and it has a powerful UI, so we support it too.

How is it implemented?
1. We have a global tracer object, that is used to create spans and pass them through the system.
2. It is a facade that hides two tracing backends with the same interface: TraceparentBackend and SentryBackend.
3. We have CherryPy tool that processes incoming requests, extracts traceparent header (or generates its parts),
   creates a span for each request, injects traceparent header into the response.
4. For each outcoming request, we ask tracer to create a new span and to inject tracing headers into the request.
   To avoid boilerplate, there is a TracedSession, an extension to requests that does all that.
   For async requests, there is a TracedAsyncSession, that does the same.
5. When the request is finished, we ask tracer to finish/pop the current span.

Logging:
There is a trace record factory, that adds a new field trace_params to each log record.
trace_params is a string representation of trace id, span id and parent span id.
If in current context they are empty (like in MainThread that doesn't process requests), trace_params is an empty string.

Sentry reporting:
If Sentry is enabled, we send info about errors and exceptions into it. We also send logs that preceded the problem
 as breadcrumbs to understand context of the error.
As we keep Sentry updated about the current trace and span, when an error happens, info about the trace will be sent to Sentry.
So we will know which chain of requests caused the error.
"""


